# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the database playground."""

from unittest import mock

import django.test
from django.conf import settings
from django.contrib.auth.hashers import check_password
from django.utils import timezone

import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianBinaryPackages,
    DebianPackageBuildLog,
    DebianSourcePackage,
    DebianUpload,
    DebusineSigningInput,
    DebusineSigningOutput,
    SigningResult,
    TaskTypes,
)
from debusine.assets import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
    AssetCategory,
    CloudProvidersType,
    DummyProviderAccountData,
    KeyPurpose,
    SigningKeyData,
)
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Asset,
    Collection,
    DEFAULT_FILE_STORE_NAME,
    FileInArtifact,
    FileStore,
    Scope,
    Token,
    User,
    WorkRequest,
)
from debusine.db.models.scopes import ScopeRole
from debusine.db.playground import scenarios
from debusine.server.scheduler import AssignToWorkerError
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.worker_pools.models import (
    AWSEC2WorkerPoolSpecification,
    DummyWorkerPoolSpecification,
    HetznerCloudWorkerPoolSpecification,
    WorkerPoolLimits,
)
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    ActionTypes,
    BackendType,
    SbuildBuildComponent,
    SbuildDynamicData,
)
from debusine.test import TestCase
from debusine.test.playground import Playground


class PlaygroundTests(django.test.TestCase, TestCase):
    """Test playground functions."""

    def assert_artifact_relations(
        self,
        artifact: Artifact,
        targets: list[tuple[Artifact, ArtifactRelation.Relations]],
    ) -> None:
        """Check that an artifact has the given set of relations."""
        actual = [
            (relation.target, ArtifactRelation.Relations(relation.type))
            for relation in ArtifactRelation.objects.filter(artifact=artifact)
        ]
        self.assertEqual(
            sorted(actual, key=lambda x: (x[0].pk, x[1])),
            sorted(targets, key=lambda x: (x[0].pk, x[1])),
        )

    def test_defaults(self) -> None:
        """Check default playground configuration."""
        with Playground() as playground:
            user = playground.get_default_user()
            self.assertEqual(user.username, "playground")

            file_store = playground.get_default_file_store()
            self.assertEqual(file_store.name, DEFAULT_FILE_STORE_NAME)
            self.assertEqual(
                file_store.backend, FileStore.BackendChoices.MEMORY
            )

            workspace = playground.get_default_workspace()
            self.assertEqual(workspace.scope.file_stores.get(), file_store)
            self.assertEqual(
                workspace.name, settings.DEBUSINE_DEFAULT_WORKSPACE
            )

    def test_user_password(self) -> None:
        """Check that the default user password is set when requested."""
        playground = Playground(
            default_username="test1", default_user_email="test1@example.org"
        )
        user = playground.get_default_user()
        self.assertEqual(user.username, "test1")
        self.assertFalse(user.has_usable_password())

        playground = Playground(
            default_username="test2",
            default_user_password="test",
            default_user_email="test2@example.org",
        )
        user = playground.get_default_user()
        self.assertEqual(user.username, "test2")
        self.assertTrue(user.has_usable_password())
        self.assertTrue(check_password("test", user.password))

    def test_create_workspace_scope(self) -> None:
        """Test setting scope on created workspaces."""
        playground = Playground()
        testscope = playground.get_or_create_scope("testscope")

        ws1 = playground.create_workspace()
        self.assertEqual(playground.create_workspace(), ws1)

        ws2 = playground.create_workspace(scope=testscope)
        self.assertEqual(playground.create_workspace(scope=testscope), ws2)
        self.assertNotEqual(ws1, ws2)

    def test_create_parent_workspace(self) -> None:
        playground = Playground()
        default_workspace = playground.get_default_workspace()

        parent = playground.create_parent_workspace(name="parent")

        self.assertIn(parent, default_workspace.inherits.all())

    def test_create_worker_without_pool(self) -> None:
        playground = Playground()
        worker = playground.create_worker()
        self.assertIsNone(worker.worker_pool)
        self.assertIsNone(worker.instance_created_at)

    def test_create_worker_with_pool(self) -> None:
        playground = Playground()
        worker_pool = playground.create_worker_pool()
        worker = playground.create_worker(worker_pool=worker_pool)
        self.assertEqual(worker.worker_pool, worker_pool)
        self.assertIsNotNone(worker.instance_created_at)

    def test_create_worker_pool_explicit(self) -> None:
        playground = Playground()
        provider_account = playground.create_cloud_provider_account_asset()
        specifications = AWSEC2WorkerPoolSpecification(launch_templates=[])
        limits = WorkerPoolLimits(max_active_instances=42)
        pool = playground.create_worker_pool(
            name="foo",
            enabled=False,
            architectures=["amd64"],
            tags=["a"],
            provider_account=provider_account,
            specifications=specifications,
            instance_wide=False,
            ephemeral=True,
            limits=limits,
        )
        self.assertEqual(pool.name, "foo")
        self.assertFalse(pool.enabled)
        self.assertEqual(pool.architectures, ["amd64"])
        self.assertEqual(pool.tags, ["a"])
        self.assertEqual(pool.provider_account, provider_account)
        self.assertEqual(pool.specifications, specifications.model_dump())
        self.assertFalse(pool.instance_wide)
        self.assertTrue(pool.ephemeral)
        self.assertEqual(pool.limits, limits.model_dump())

    def test_create_worker_pool_defaults(self) -> None:
        playground = Playground()
        pool = playground.create_worker_pool()
        self.assertEqual(pool.name, "test")
        self.assertTrue(pool.enabled)
        self.assertEqual(pool.architectures, ["amd64", "i386"])
        self.assertEqual(pool.tags, [])
        self.assertIsNotNone(pool.provider_account, Asset)
        self.assertEqual(
            pool.specifications, DummyWorkerPoolSpecification().model_dump()
        )
        self.assertTrue(pool.instance_wide)
        self.assertFalse(pool.ephemeral)
        self.assertEqual(pool.limits, WorkerPoolLimits().model_dump())

    def test_create_worker_pool_aws(self) -> None:
        playground = Playground()
        pool = playground.create_worker_pool(
            provider_account=playground.create_cloud_provider_account_asset(
                cloud_provider=CloudProvidersType.AWS
            ),
        )
        self.assertEqual(
            pool.specifications,
            AWSEC2WorkerPoolSpecification(launch_templates=[]).model_dump(),
        )

    def test_create_worker_pool_hetzner(self) -> None:
        playground = Playground()
        pool = playground.create_worker_pool(
            provider_account=playground.create_cloud_provider_account_asset(
                cloud_provider=CloudProvidersType.HETZNER
            ),
        )
        self.assertEqual(
            pool.specifications,
            HetznerCloudWorkerPoolSpecification(
                server_type="cx22",
                image_name="debian-13",
            ).model_dump(),
        )

    def test_create_worker_pool_unknown_provider(self) -> None:
        playground = Playground()
        provider_account = playground.create_cloud_provider_account_asset()
        provider_account.data = {"provider_type": "unknown"}
        with context.disable_permission_checks():
            provider_account.save()

        with self.assertRaisesRegex(
            NotImplementedError, r"No support for unknown"
        ):
            playground.create_worker_pool(provider_account=provider_account)

    def test_create_file_in_backend(self) -> None:
        """create_file_in_backend() returns a File and writes contents to it."""
        playground = self.enterContext(Playground(memory_file_store=False))
        file_backend = FileStore.default().get_backend_object()
        contents = b"some-test-data"

        fileobj = playground.create_file_in_backend(file_backend, contents)

        self.assertIsInstance(fileobj.id, int)
        local_path = file_backend.get_local_path(fileobj)
        assert local_path is not None
        self.assertEqual(local_path.read_bytes(), contents)

    def test_create_source_artifact(self) -> None:
        """Test creating a source artifact."""
        playground = Playground()
        source = playground.create_source_artifact()
        self.assertEqual(source.category, ArtifactCategory.SOURCE_PACKAGE)
        self.assertEqual(source.workspace, playground.get_default_workspace())
        self.assertEqual(source.files.count(), 0)
        self.assertEqual(source.created_by, playground.get_default_user())
        self.assertIsNone(source.created_by_work_request)

        artifact = DebianSourcePackage(**source.data)
        self.assert_source_artifact_equal(artifact, "hello", "1.0-1")

    def test_create_minimal_binary_packages_artifact(self) -> None:
        """Test creating a binary_packages artifact."""
        playground = Playground()
        bp = playground.create_minimal_binary_packages_artifact(
            "hello", "1.0-1", "1.0-1", "amd64"
        )
        self.assertEqual(bp.category, ArtifactCategory.BINARY_PACKAGES)
        self.assertEqual(bp.workspace, playground.get_default_workspace())
        self.assertEqual(bp.files.count(), 0)
        self.assertIsNone(bp.created_by)
        self.assertIsNone(bp.created_by_work_request)

        artifact = DebianBinaryPackages(**bp.data)
        self.assertEqual(artifact.srcpkg_name, "hello")
        self.assertEqual(artifact.srcpkg_version, "1.0-1")
        self.assertEqual(artifact.version, "1.0-1")
        self.assertEqual(artifact.architecture, "amd64")
        self.assertEqual(artifact.packages, [])

    def test_create_minimal_binary_package_artifact(self) -> None:
        """Test creating a binary_package artifact."""
        playground = Playground()
        bp = playground.create_minimal_binary_package_artifact(
            "hello", "1.0-1", "hello", "1.0-1", "amd64"
        )
        self.assertEqual(bp.category, ArtifactCategory.BINARY_PACKAGE)
        self.assertEqual(bp.workspace, playground.get_default_workspace())
        self.assertEqual(bp.files.count(), 0)
        self.assertIsNone(bp.created_by)
        self.assertIsNone(bp.created_by_work_request)

        artifact = DebianBinaryPackage(**bp.data)
        self.assertEqual(artifact.srcpkg_name, "hello")
        self.assertEqual(artifact.srcpkg_version, "1.0-1")
        self.assertEqual(artifact.deb_control_files, [])
        self.assertEqual(
            artifact.deb_fields,
            {
                "Architecture": "amd64",
                "Package": "hello",
                "Version": "1.0-1",
            },
        )

    def test_create_source_artifact_with_files(self) -> None:
        """Test creating a source artifact with its files."""
        playground = Playground()
        source = playground.create_source_artifact(create_files=True)
        files = sorted(
            FileInArtifact.objects.filter(artifact=source),
            key=lambda f: f.path,
        )
        self.assertEqual(len(files), 3)
        self.assertEqual(files[0].path, "hello_1.0-1.debian.tar.xz")
        self.assertEqual(files[1].path, "hello_1.0-1.dsc")
        self.assertEqual(files[2].path, "hello_1.0.orig.tar.gz")

    def test_create_upload_artifacts(self) -> None:
        """Test creating a set of upload artifacts."""
        playground = Playground()
        upload_artifacts = playground.create_upload_artifacts(
            binaries=[("bin1", "all")]
        )
        upload = upload_artifacts.upload
        self.assertEqual(upload.category, ArtifactCategory.UPLOAD)
        self.assertEqual(upload.workspace, playground.get_default_workspace())
        self.assertEqual(upload.files.count(), 0)
        self.assertEqual(upload.created_by, playground.get_default_user())
        self.assertIsNone(upload.created_by_work_request)

        DebianUpload(**upload.data)

        source = upload_artifacts.source
        self.assertIsNotNone(source)
        self.assertEqual(source.category, ArtifactCategory.SOURCE_PACKAGE)
        self.assertEqual(source.workspace, playground.get_default_workspace())
        self.assertEqual(source.files.count(), 0)
        self.assertEqual(source.created_by, playground.get_default_user())
        self.assertIsNone(source.created_by_work_request)
        self.assertTrue(
            ArtifactRelation.objects.filter(
                artifact=upload,
                target=source,
                type=ArtifactRelation.Relations.EXTENDS,
            ).exists()
        )

        binaries = upload_artifacts.binaries
        self.assertIsNotNone(binaries)
        self.assertEqual(len(binaries), 1)
        binary = binaries[0]
        self.assertEqual(binary.category, ArtifactCategory.BINARY_PACKAGE)
        self.assertEqual(binary.workspace, playground.get_default_workspace())
        self.assertEqual(binary.files.count(), 0)
        self.assertEqual(binary.created_by, playground.get_default_user())
        self.assertIsNone(binary.created_by_work_request)

        self.assertTrue(
            ArtifactRelation.objects.filter(
                artifact=upload,
                target=binary,
                type=ArtifactRelation.Relations.EXTENDS,
            ).exists()
        )

    def test_create_upload_artifacts_no_binaries(self) -> None:
        """Test creating a set of upload source artifacts."""
        playground = Playground()
        upload_artifacts = playground.create_upload_artifacts(binary=False)
        upload = upload_artifacts.upload
        self.assertEqual(upload.category, ArtifactCategory.UPLOAD)
        self.assertEqual(upload.workspace, playground.get_default_workspace())
        self.assertEqual(upload.files.count(), 0)
        self.assertEqual(upload.created_by, playground.get_default_user())
        self.assertIsNone(upload.created_by_work_request)

        DebianUpload(**upload.data)

        source = upload_artifacts.source
        self.assertIsNotNone(source)
        self.assertEqual(source.category, ArtifactCategory.SOURCE_PACKAGE)
        self.assertEqual(source.workspace, playground.get_default_workspace())
        self.assertEqual(source.files.count(), 0)
        self.assertEqual(source.created_by, playground.get_default_user())
        self.assertIsNone(source.created_by_work_request)
        self.assertTrue(
            ArtifactRelation.objects.filter(
                artifact=upload,
                target=source,
                type=ArtifactRelation.Relations.EXTENDS,
            ).exists()
        )

        self.assertIsNone(upload_artifacts.binaries)

    def test_create_upload_artifacts_binary_only(self) -> None:
        """Test creating a set of upload binary artifacts."""
        playground = Playground()
        upload_artifacts = playground.create_upload_artifacts(
            source=False, binaries=[("foo", "all")]
        )
        upload = upload_artifacts.upload
        self.assertEqual(upload.category, ArtifactCategory.UPLOAD)
        self.assertEqual(upload.workspace, playground.get_default_workspace())
        self.assertEqual(upload.files.count(), 0)
        self.assertEqual(upload.created_by, playground.get_default_user())
        self.assertIsNone(upload.created_by_work_request)

        DebianUpload(**upload.data)

        self.assertIsNone(upload_artifacts.source)

        binaries = upload_artifacts.binaries
        self.assertIsNotNone(binaries)
        self.assertEqual(len(binaries), 1)
        binary = binaries[0]
        self.assertEqual(binary.category, ArtifactCategory.BINARY_PACKAGE)
        self.assertEqual(binary.workspace, playground.get_default_workspace())
        self.assertEqual(binary.files.count(), 0)
        self.assertEqual(binary.created_by, playground.get_default_user())
        self.assertIsNone(binary.created_by_work_request)

        self.assertTrue(
            ArtifactRelation.objects.filter(
                artifact=upload,
                target=binary,
                type=ArtifactRelation.Relations.EXTENDS,
            ).exists()
        )

    def test_create_build_log_artifact(self) -> None:
        """Test creating a build log artifact."""
        playground = Playground()
        buildlog = playground.create_build_log_artifact()
        self.assertEqual(buildlog.category, ArtifactCategory.PACKAGE_BUILD_LOG)
        self.assertEqual(buildlog.workspace, playground.get_default_workspace())
        self.assertEqual(buildlog.files.count(), 1)
        self.assertEqual(buildlog.created_by, playground.get_default_user())
        self.assertIsNone(buildlog.created_by_work_request)

        artifact = DebianPackageBuildLog(**buildlog.data)
        self.assertEqual(artifact.source, "hello")
        self.assertEqual(artifact.version, "1.0-1")
        self.assertEqual(artifact.filename, "hello_1.0-1_amd64.buildlog")

        file = buildlog.files.first()
        assert file is not None
        file_backend = buildlog.workspace.scope.download_file_backend(file)
        with file_backend.get_stream(file) as fd:
            self.assertEqual(
                fd.read().splitlines(keepends=True)[3].decode(),
                "Line 4 of hello_1.0-1_amd64.buildlog\n",
            )

    def test_create_build_log_artifact_custom(self) -> None:
        """Test creating a build log artifact with custom arguments."""
        test_contents = b"test contents"
        playground = Playground()
        user = User.objects.create_user(
            username="custom", email="custom@example.org"
        )
        work_request = playground.create_worker_task(created_by=user)
        buildlog = playground.create_build_log_artifact(
            source="test",
            version="2.0",
            build_arch="arm64",
            work_request=work_request,
            contents=test_contents,
        )
        self.assertEqual(buildlog.category, ArtifactCategory.PACKAGE_BUILD_LOG)
        self.assertEqual(buildlog.workspace, playground.get_default_workspace())
        self.assertEqual(buildlog.files.count(), 1)
        self.assertEqual(buildlog.created_by, user)
        self.assertEqual(buildlog.created_by_work_request, work_request)

        artifact = DebianPackageBuildLog(**buildlog.data)
        self.assertEqual(artifact.source, "test")
        self.assertEqual(artifact.version, "2.0")
        self.assertEqual(artifact.filename, "test_2.0_arm64.buildlog")

        file = buildlog.files.first()
        assert file is not None
        file_backend = buildlog.workspace.scope.download_file_backend(file)
        with file_backend.get_stream(file) as fd:
            self.assertEqual(fd.read(), test_contents)

    def test_create_build_log_artifact_custom_user(self) -> None:
        """Test creating a build log artifact with custom user."""
        playground = Playground()
        user = User.objects.create_user(
            username="custom", email="custom@example.org"
        )
        work_request = playground.create_worker_task()
        buildlog = playground.create_build_log_artifact(
            work_request=work_request,
            created_by=user,
        )
        self.assertEqual(buildlog.created_by, user)
        self.assertEqual(buildlog.created_by_work_request, work_request)

    def test_create_signing_input_artifact(self) -> None:
        playground = Playground()
        bp = playground.create_signing_input_artifact("hello")
        self.assertEqual(bp.category, ArtifactCategory.SIGNING_INPUT)
        self.assertEqual(bp.workspace, playground.get_default_workspace())
        self.assertEqual(bp.files.count(), 0)
        self.assertIsNone(bp.created_by)
        self.assertIsNone(bp.created_by_work_request)

        artifact = DebusineSigningInput(**bp.data)
        self.assertEqual(artifact.binary_package_name, "hello")
        self.assertIsNone(artifact.trusted_certs)

    def test_create_signing_output_artifact(self) -> None:
        playground = Playground()
        bp = playground.create_signing_output_artifact(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABC123",
            results=[SigningResult(file="test", output_file="test.sig")],
            binary_package_name="hello",
        )
        self.assertEqual(bp.category, ArtifactCategory.SIGNING_OUTPUT)
        self.assertEqual(bp.workspace, playground.get_default_workspace())
        self.assertEqual(bp.files.count(), 0)
        self.assertIsNone(bp.created_by)
        self.assertIsNone(bp.created_by_work_request)

        artifact = DebusineSigningOutput(**bp.data)
        self.assertEqual(artifact.purpose, KeyPurpose.UEFI)
        self.assertEqual(artifact.fingerprint, "ABC123")
        self.assertEqual(
            artifact.results,
            [SigningResult(file="test", output_file="test.sig")],
        )
        self.assertEqual(artifact.binary_package_name, "hello")

    def test_create_signing_key_asset_defaults(self) -> None:
        """Test create_signing_key_asset with defaults."""
        playground = Playground()
        asset = playground.create_signing_key_asset()
        self.assertEqual(asset.category, AssetCategory.SIGNING_KEY)
        self.assertEqual(asset.workspace, playground.get_default_workspace())
        self.assertEqual(asset.created_by, playground.get_default_user())
        self.assertIsNone(asset.created_by_work_request)
        data_model = asset.data_model
        assert isinstance(data_model, SigningKeyData)
        self.assertEqual(data_model.purpose, KeyPurpose.OPENPGP_UPLOAD)
        self.assertGreater(len(data_model.fingerprint), 8)
        self.assertTrue(
            data_model.public_key.startswith(
                b"-----BEGIN PGP PUBLIC KEY BLOCK-----"
            )
        )
        self.assertEqual(data_model.description, "Test Key")

    def test_create_signing_key_asset_custom(self) -> None:
        """Test create_signing_key_asset with specified features."""
        playground = Playground()
        workspace = playground.create_workspace()
        user = playground.create_user(username="test")
        asset = playground.create_signing_key_asset(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABCDEF",
            public_key="PUBLIC KEY",
            description="Custom Asset",
            workspace=workspace,
            created_by=user,
        )
        self.assertEqual(asset.workspace, workspace)
        self.assertEqual(asset.created_by, user)
        self.assertIsNone(asset.created_by_work_request)
        data_model = asset.data_model
        assert isinstance(data_model, SigningKeyData)
        self.assertEqual(data_model.purpose, KeyPurpose.UEFI)
        self.assertEqual(data_model.fingerprint, "ABCDEF")
        self.assertEqual(data_model.public_key, b"PUBLIC KEY")
        self.assertEqual(data_model.description, "Custom Asset")

    def test_create_cloud_provider_account_asset_defaults(self) -> None:
        playground = Playground()
        asset = playground.create_cloud_provider_account_asset()
        self.assertEqual(
            asset.data, DummyProviderAccountData(name="test").model_dump()
        )

    def test_create_cloud_provider_account_asset_custom_data(self) -> None:
        playground = Playground()
        data = DummyProviderAccountData(name="test", secret="custom-secret")
        asset = playground.create_cloud_provider_account_asset(data=data)
        self.assertEqual(asset.data, data.model_dump())

    def test_create_cloud_provider_account_asset_aws(self) -> None:
        playground = Playground()
        asset = playground.create_cloud_provider_account_asset(
            name="aws-test", cloud_provider=CloudProvidersType.AWS
        )
        self.assertEqual(
            asset.data,
            AWSProviderAccountData(
                name="aws-test",
                configuration=AWSProviderAccountConfiguration(
                    region_name="test-region"
                ),
                credentials=AWSProviderAccountCredentials(
                    access_key_id="access-key",
                    secret_access_key="secret-key",
                ),
            ).model_dump(),
        )

    def test_create_asset_usage_defaults(self) -> None:
        """Test create_asset_usage with default workspace."""
        playground = Playground()
        asset = playground.create_signing_key_asset()
        asset_usage = playground.create_asset_usage(resource=asset)
        self.assertEqual(asset_usage.asset, asset)
        self.assertEqual(
            asset_usage.workspace, playground.get_default_workspace()
        )

    def test_create_asset_usage_custom(self) -> None:
        """Test create_asset_usage with specified workspace."""
        playground = Playground()
        asset = playground.create_signing_key_asset()
        workspace = playground.create_workspace()
        asset_usage = playground.create_asset_usage(
            resource=asset, workspace=workspace
        )
        self.assertEqual(asset_usage.asset, asset)
        self.assertEqual(asset_usage.workspace, workspace)

    def test_create_debian_env_collection_defaults(self) -> None:
        """Test create_debian_environments_collection."""
        playground = Playground()
        env = playground.create_debian_environments_collection()
        self.assertEqual(env.name, "debian")
        self.assertEqual(env.category, CollectionCategory.ENVIRONMENTS)
        self.assertEqual(env.workspace, playground.get_default_workspace())

    def test_create_debian_env_collection_custom(self) -> None:
        """Test create_debian_environments_collection."""
        playground = Playground()
        workspace = playground.create_workspace(name="custom")
        env = playground.create_debian_environments_collection(
            name="ubuntu", workspace=workspace
        )
        self.assertEqual(env.name, "ubuntu")
        self.assertEqual(env.category, CollectionCategory.ENVIRONMENTS)
        self.assertEqual(env.workspace, workspace)

    def test_create_debian_env_defaults(self) -> None:
        """Test create_debian_environment."""
        playground = Playground()
        env_item = playground.create_debian_environment()
        self.assertEqual(env_item.category, ArtifactCategory.SYSTEM_TARBALL)
        self.assertIsNotNone(env_item.artifact)
        self.assertEqual(
            env_item.parent_collection,
            playground.create_debian_environments_collection(),
        )
        self.assertEqual(
            env_item.created_by_user, playground.get_default_user()
        )
        self.assertEqual(
            env_item.data,
            {
                'architecture': 'amd64',
                'backend': BackendType.UNSHARE,
                'codename': 'bookworm',
                'variant': None,
            },
        )

        env = env_item.artifact
        assert env is not None
        self.assertEqual(env.category, ArtifactCategory.SYSTEM_TARBALL)
        self.assertEqual(env.workspace, playground.get_default_workspace())
        self.assertEqual(
            env.data,
            {
                "architecture": "amd64",
                "codename": "bookworm",
                "components": ["main"],
                'filename': 'test',
                'mirror': 'https://deb.debian.org',
                'pkglist': {},
                'variant': None,
                'vendor': 'Debian',
                "with_dev": True,
                'with_init': True,
            },
        )

    def test_create_debian_env_custom(self) -> None:
        """Test create_debian_environment with custom args."""
        playground = Playground()
        workspace = playground.create_workspace(name="custom")
        collection = playground.create_debian_environments_collection(
            workspace=workspace
        )
        user = User.objects.create_user(
            username="custom", email="custom@example.org"
        )
        env_item = playground.create_debian_environment(
            workspace=workspace, variant="apt", collection=collection, user=user
        )
        self.assertEqual(env_item.category, ArtifactCategory.SYSTEM_TARBALL)
        self.assertIsNotNone(env_item.artifact)
        self.assertEqual(env_item.parent_collection, collection)
        self.assertEqual(env_item.created_by_user, user)
        self.assertEqual(
            env_item.data,
            {
                'architecture': 'amd64',
                'backend': BackendType.UNSHARE,
                'codename': 'bookworm',
                'variant': 'apt',
            },
        )

        env = env_item.artifact
        assert env is not None
        self.assertEqual(env.category, ArtifactCategory.SYSTEM_TARBALL)
        self.assertEqual(env.workspace, workspace)
        self.assertEqual(
            env.data,
            {
                "architecture": "amd64",
                "codename": "bookworm",
                "components": ["main"],
                'filename': 'test',
                'mirror': 'https://deb.debian.org',
                'pkglist': {},
                "variant": "apt",
                'vendor': 'Debian',
                "with_dev": True,
                'with_init': True,
            },
        )

    def test_create_debian_env_reuse(self) -> None:
        """Test object reuse of create_debian_environment."""
        playground = Playground()
        env_item1 = playground.create_debian_environment()
        env_item2 = playground.create_debian_environment()
        self.assertEqual(env_item1, env_item2)

        env_item2 = playground.create_debian_environment(
            environment=env_item1.artifact
        )
        self.assertEqual(env_item1, env_item2)

    def test_create_debian_env_image(self) -> None:
        """Test create_debian_environment for images."""
        playground = Playground()
        env_item = playground.create_debian_environment(
            category=ArtifactCategory.SYSTEM_IMAGE
        )
        self.assertEqual(env_item.category, ArtifactCategory.SYSTEM_IMAGE)
        self.assertIsNotNone(env_item.artifact)
        self.assertEqual(
            env_item.parent_collection,
            playground.create_debian_environments_collection(),
        )
        self.assertEqual(
            env_item.data,
            {
                'architecture': 'amd64',
                'backend': BackendType.UNSHARE,
                'codename': 'bookworm',
                'variant': None,
            },
        )

        env = env_item.artifact
        assert env is not None
        self.assertEqual(env.category, ArtifactCategory.SYSTEM_IMAGE)
        self.assertEqual(env.workspace, playground.get_default_workspace())
        self.assertEqual(
            env.data,
            {
                "architecture": "amd64",
                "codename": "bookworm",
                "components": ["main"],
                'filename': 'test',
                'mirror': 'https://deb.debian.org',
                'pkglist': {},
                'variant': None,
                'vendor': 'Debian',
                "with_dev": True,
                'with_init': True,
            },
        )

    def test_create_sbuild_work_request(self) -> None:
        """Test creating a sbuild work request."""
        playground = Playground()
        source = playground.create_source_artifact()
        environment_item = playground.create_debian_environment()
        assert environment_item.artifact is not None
        wr = playground.create_sbuild_work_request(
            source=source,
            environment=environment_item.artifact,
            architecture="amd64",
        )

        self.assertEqual(wr.workspace, playground.get_default_workspace())
        self.assertEqual(wr.created_by, playground.get_default_user())
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)
        self.assertIsNone(wr.worker)
        self.assertEqual(wr.task_type, TaskTypes.WORKER)
        self.assertEqual(wr.task_name, "sbuild")
        self.assertEqual(
            wr.task_data,
            {
                'backend': BackendType.UNSHARE,
                'build_components': [SbuildBuildComponent.ANY],
                'environment': environment_item.artifact.pk,
                'build_architecture': 'amd64',
                'input': {'source_artifact': source.pk},
            },
        )
        self.assertIsNone(wr.dynamic_task_data)
        self.assertIsNone(wr.parent)

    def test_simulate_package_build(self) -> None:
        """Test simulating a whole package build."""
        playground = Playground()
        source = playground.create_source_artifact()
        wr = playground.simulate_package_build(source, architecture="amd64")
        task_config_collection = Collection.objects.get(
            workspace=playground.get_default_workspace(),
            name="default",
            category=CollectionCategory.TASK_CONFIGURATION,
        )

        binaries: list[Artifact] = []
        buildlogs: list[Artifact] = []
        uploads: list[Artifact] = []
        for artifact in Artifact.objects.filter(created_by_work_request=wr):
            match artifact.category:
                case ArtifactCategory.BINARY_PACKAGE:
                    binaries.append(artifact)
                case ArtifactCategory.PACKAGE_BUILD_LOG:
                    buildlogs.append(artifact)
                case ArtifactCategory.UPLOAD:
                    uploads.append(artifact)
                case _ as unreachable:
                    self.fail(
                        "Work request generated unexpected"
                        f" {unreachable} artifact"
                    )
        self.assertEqual(len(binaries), 1)
        self.assertEqual(len(buildlogs), 1)
        self.assertEqual(len(uploads), 1)

        self.assertEqual(wr.workspace, playground.get_default_workspace())
        self.assertIsNotNone(wr.started_at)
        self.assertIsNotNone(wr.completed_at)
        self.assertEqual(wr.created_by, playground.get_default_user())
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)
        self.assertIsNotNone(wr.worker)
        self.assertEqual(wr.task_type, TaskTypes.WORKER)
        self.assertEqual(wr.task_name, "sbuild")
        environment = Artifact.objects.get(pk=wr.task_data["environment"])
        source = Artifact.objects.get(
            pk=wr.task_data["input"]["source_artifact"]
        )
        task_data = {
            'backend': BackendType.UNSHARE,
            'build_components': [SbuildBuildComponent.ANY],
            'environment': environment.pk,
            'build_architecture': 'amd64',
            'input': {'source_artifact': source.pk},
        }
        assert wr.task_data == task_data
        self.assertEqual(
            wr.configured_task_data,
            task_data | {"task_configuration": task_config_collection.id},
        )
        assert wr.dynamic_task_data is not None
        dynamic_data = wr.dynamic_task_data.copy()
        task_input_fields = dynamic_data.pop("task_input_fields")
        self.assertEqual(
            SbuildDynamicData.model_validate(dynamic_data),
            SbuildDynamicData(
                environment_id=environment.pk,
                input_source_artifact_id=source.pk,
                binnmu_maintainer=(
                    f"Debusine <noreply@{settings.DEBUSINE_FQDN}>"
                ),
                subject="hello",
                parameter_summary="hello_1.0-1 (amd64)",
                runtime_context="any:amd64:amd64",
                configuration_context="bookworm",
            ),
        )
        self.assertCountEqual(
            task_input_fields.keys(),
            [
                "build_architecture",
                "debusine_fqdn",
                "environment",
                "extra_repositories",
                "input_extra_binary_artifacts",
                "input_source_artifact",
                "task_configuration",
            ],
        )
        self.assertCountEqual(
            wr.scheduler_tags_required,
            [
                wtags.TASK_PREFIX + "worker:sbuild:version:1",
                wtags.WORKER_TYPE_EXTERNAL,
                wtags.EXECUTOR_PREFIX + "unshare",
                wtags.CAP_PREFIX + "sbuild",
                wtags.BUILD_ARCH_PREFIX + "amd64",
            ],
        )
        self.assertIsNone(wr.parent)
        self.assertEqual(wr.event_reactions_json, {})

        # Test environment
        self.assertEqual(environment.category, ArtifactCategory.SYSTEM_TARBALL)

        # Test source
        self.assertEqual(source.category, ArtifactCategory.SOURCE_PACKAGE)
        self.assert_source_artifact_equal(
            DebianSourcePackage(**source.data), name="hello", version="1.0-1"
        )

        # Test buildlog
        buildlog = buildlogs[0]
        self.assertEqual(buildlog.category, ArtifactCategory.PACKAGE_BUILD_LOG)
        self.assertEqual(
            buildlog.data,
            {
                'filename': 'hello_1.0-1_amd64.buildlog',
                'source': 'hello',
                'version': '1.0-1',
                'architecture': 'amd64',
                'bd_uninstallable': None,
            },
        )

        # Test binary
        binary = binaries[0]
        self.assertEqual(binary.category, ArtifactCategory.BINARY_PACKAGE)
        self.assertEqual(
            binary.data,
            {
                'deb_control_files': ['control'],
                'deb_fields': {
                    'Architecture': 'amd64',
                    'Description': 'Example description',
                    'Maintainer': 'Example Maintainer <example@example.org>',
                    'Package': 'hello',
                    'Priority': 'optional',
                    'Section': 'unknown',
                    'Version': '1.0-1',
                },
                'srcpkg_name': 'hello',
                'srcpkg_version': '1.0-1',
            },
        )

        # Test upload
        upload = uploads[0]
        self.assertEqual(upload.category, ArtifactCategory.UPLOAD)
        self.assertEqual(upload.data["changes_fields"]["Architecture"], "amd64")
        self.assertEqual(upload.data["changes_fields"]["Binary"], "hello")
        self.assertEqual(upload.data["changes_fields"]["Source"], "hello")
        self.assertEqual(upload.data["changes_fields"]["Version"], "1.0-1")

        # Test artifact relations
        self.assert_artifact_relations(
            buildlog,
            [
                (source, ArtifactRelation.Relations.RELATES_TO),
                (binary, ArtifactRelation.Relations.RELATES_TO),
                (upload, ArtifactRelation.Relations.RELATES_TO),
            ],
        )
        self.assert_artifact_relations(source, [])
        self.assert_artifact_relations(
            binary, [(source, ArtifactRelation.Relations.BUILT_USING)]
        )
        self.assert_artifact_relations(
            upload, [(binary, ArtifactRelation.Relations.EXTENDS)]
        )

    def test_simulate_package_build_custom(self) -> None:
        """Test simulating a whole package build with custom args."""
        playground = Playground()

        workspace = playground.create_workspace(name="custom", public=True)
        source = playground.create_source_artifact(
            name="test", version="2.0", workspace=workspace
        )
        env_item = playground.create_debian_environment(
            workspace=workspace,
        )
        environment = env_item.artifact
        assert environment is not None
        worker = playground.create_worker()

        wr = playground.simulate_package_build(
            source,
            environment=environment,
            worker=worker,
            architecture="amd64",
        )

        binaries: list[Artifact] = []
        buildlogs: list[Artifact] = []
        uploads: list[Artifact] = []
        for artifact in Artifact.objects.filter(created_by_work_request=wr):
            match artifact.category:
                case ArtifactCategory.BINARY_PACKAGE:
                    binaries.append(artifact)
                case ArtifactCategory.PACKAGE_BUILD_LOG:
                    buildlogs.append(artifact)
                case ArtifactCategory.UPLOAD:
                    uploads.append(artifact)
                case _ as unreachable:
                    self.fail(
                        "Work request generated unexpected"
                        f" {unreachable} artifact"
                    )
        self.assertEqual(len(binaries), 1)
        self.assertEqual(len(buildlogs), 1)

        self.assertEqual(wr.workspace, workspace)
        self.assertIsNotNone(wr.started_at)
        self.assertIsNotNone(wr.completed_at)
        self.assertEqual(wr.created_by, playground.get_default_user())
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)
        self.assertEqual(wr.worker, worker)
        self.assertEqual(wr.task_type, TaskTypes.WORKER)
        self.assertEqual(wr.task_name, "sbuild")
        self.assertEqual(
            Artifact.objects.get(pk=wr.task_data["environment"]), environment
        )
        self.assertEqual(
            Artifact.objects.get(pk=wr.task_data["input"]["source_artifact"]),
            source,
        )
        self.assertEqual(
            wr.task_data,
            {
                'backend': BackendType.UNSHARE,
                'build_components': [SbuildBuildComponent.ANY],
                'environment': environment.pk,
                'build_architecture': 'amd64',
                'input': {'source_artifact': source.pk},
            },
        )
        assert wr.dynamic_task_data is not None
        dynamic_data = wr.dynamic_task_data.copy()
        task_input_fields = dynamic_data.pop("task_input_fields")
        self.assertEqual(
            SbuildDynamicData.model_validate(dynamic_data),
            SbuildDynamicData(
                environment_id=environment.pk,
                input_source_artifact_id=source.pk,
                binnmu_maintainer=(
                    f"Debusine <noreply@{settings.DEBUSINE_FQDN}>"
                ),
                subject="test",
                parameter_summary="test_2.0 (amd64)",
                runtime_context="any:amd64:amd64",
                configuration_context="bookworm",
            ),
        )
        self.assertCountEqual(
            task_input_fields.keys(),
            [
                "build_architecture",
                "debusine_fqdn",
                "environment",
                "extra_repositories",
                "input_extra_binary_artifacts",
                "input_source_artifact",
                "task_configuration",
            ],
        )
        self.assertCountEqual(
            wr.scheduler_tags_required,
            [
                wtags.TASK_PREFIX + "worker:sbuild:version:1",
                wtags.WORKER_TYPE_EXTERNAL,
                wtags.EXECUTOR_PREFIX + "unshare",
                wtags.CAP_PREFIX + "sbuild",
                wtags.BUILD_ARCH_PREFIX + "amd64",
            ],
        )
        self.assertIsNone(wr.parent)
        self.assertEqual(wr.event_reactions_json, {})

        # Test environment
        self.assertEqual(environment.category, ArtifactCategory.SYSTEM_TARBALL)

        # Test buildlog
        buildlog = buildlogs[0]
        self.assertEqual(buildlog.category, ArtifactCategory.PACKAGE_BUILD_LOG)
        self.assertEqual(
            buildlog.data,
            {
                'filename': 'test_2.0_amd64.buildlog',
                'source': 'test',
                'version': '2.0',
                'architecture': 'amd64',
                'bd_uninstallable': None,
            },
        )

        # Test binary
        binary = binaries[0]
        self.assertEqual(binary.category, ArtifactCategory.BINARY_PACKAGE)
        self.assertEqual(
            binary.data,
            {
                'deb_control_files': ['control'],
                'deb_fields': {
                    'Architecture': 'amd64',
                    'Description': 'Example description',
                    'Maintainer': 'Example Maintainer <example@example.org>',
                    'Package': 'test',
                    'Priority': 'optional',
                    'Section': 'unknown',
                    'Version': '2.0',
                },
                'srcpkg_name': 'test',
                'srcpkg_version': '2.0',
            },
        )

        # Test upload
        upload = uploads[0]
        self.assertEqual(upload.category, ArtifactCategory.UPLOAD)
        self.assertEqual(upload.data["changes_fields"]["Architecture"], "amd64")
        self.assertEqual(upload.data["changes_fields"]["Binary"], "test")
        self.assertEqual(upload.data["changes_fields"]["Source"], "test")
        self.assertEqual(upload.data["changes_fields"]["Version"], "2.0")

        # Test artifact relations
        self.assert_artifact_relations(
            buildlog,
            [
                (source, ArtifactRelation.Relations.RELATES_TO),
                (binary, ArtifactRelation.Relations.RELATES_TO),
                (upload, ArtifactRelation.Relations.RELATES_TO),
            ],
        )
        self.assert_artifact_relations(source, [])
        self.assert_artifact_relations(
            binary, [(source, ArtifactRelation.Relations.BUILT_USING)]
        )
        self.assert_artifact_relations(
            upload, [(binary, ArtifactRelation.Relations.EXTENDS)]
        )

    def test_simulate_package_build_workflow(self) -> None:
        """Test simulating a package build in a workflow."""
        playground = Playground()
        template = playground.create_workflow_template("test", "noop")
        workflow = playground.create_workflow(template, task_data={})
        source = playground.create_source_artifact()
        wr = playground.simulate_package_build(source, workflow=workflow)
        self.assertEqual(wr.parent, workflow)
        self.assertEqual(
            wr.workflow_data,
            WorkRequestWorkflowData(
                display_name="Build all",
                step="build-all",
            ),
        )
        self.assertEqual(
            wr.event_reactions_json,
            {
                "on_success": [
                    {
                        "action": ActionTypes.UPDATE_COLLECTION_WITH_ARTIFACTS,
                        "artifact_filters": {
                            "category": ArtifactCategory.UPLOAD
                        },
                        "collection": "internal@collections",
                        "name_template": "build-all",
                        "variables": {
                            "architecture": "all",
                            "binary_names": ["hello"],
                            "source_package_name": "hello",
                        },
                    }
                ],
            },
        )

    def test_create_workflow(self) -> None:
        """Test create_workflow."""
        playground = Playground()
        workflow = playground.create_workflow()

        self.assertTrue(workflow.is_workflow)

    def test_create_work_request(self) -> None:
        """Test create_work_request return a saved work request."""
        playground = Playground()
        work_request = playground.create_worker_task()
        work_request.refresh_from_db()
        self.assertIsInstance(work_request, WorkRequest)
        self.assertEqual(
            work_request.workspace, playground.get_default_workspace()
        )

    def test_create_work_request_use_created_by_user(self) -> None:
        """Test create_work_request use created_by Token."""
        playground = Playground()
        user = playground.create_user("testuser")
        work_request = playground.create_worker_task(created_by=user)
        work_request.refresh_from_db()
        self.assertEqual(work_request.created_by, user)

    def test_create_work_request_expired(self) -> None:
        """Test create_work_request expired=True."""
        playground = Playground()
        work_request = playground.create_worker_task(expired=True)

        self.assertIsNotNone(work_request.expire_at)
        assert work_request.expire_at is not None
        self.assertLess(work_request.expire_at, timezone.now())

    def test_create_user_token(self) -> None:
        """create_user_token returns an enabled token with user."""
        playground = Playground()
        token = playground.create_user_token()
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.USER)
        self.assertEqual(token.user, playground.get_default_user())
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_user_token_with_user(self) -> None:
        """Test create_user_token with a given user."""
        playground = Playground()
        user = playground.create_user("test")
        token = playground.create_user_token(user=user)
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.USER)
        self.assertEqual(token.user, user)
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_user_token_disabled(self) -> None:
        """create_user_token returns a disabled token with user."""
        playground = Playground()
        token = playground.create_user_token(enabled=False)
        self.assertFalse(token.enabled)
        self.assertEqual(token.token_type, Token.Types.USER)
        self.assertIsNotNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_external_user_token(self) -> None:
        playground = Playground()
        token = playground.create_external_user_token(username="test")
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.EXTERNAL_USER)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "test")
        self.assertFalse(hasattr(token, "worker"))
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertTrue(token.explicit_grants_only)

    def test_create_external_user_token_disabled(self) -> None:
        playground = Playground()
        token = playground.create_external_user_token(
            username="test-disabled",
            enabled=False,
        )
        self.assertFalse(token.enabled)
        self.assertEqual(token.token_type, Token.Types.EXTERNAL_USER)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "test-disabled")
        self.assertFalse(hasattr(token, "worker"))
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertTrue(token.explicit_grants_only)

    def test_create_worker_token(self) -> None:
        """create_worker_token returns an enabled token."""
        playground = Playground()
        token = playground.create_worker_token()
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertIsNotNone(token.worker)
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_worker_token_with_worker(self) -> None:
        """create_worker_token returns a token with a worker."""
        playground = Playground()
        worker = playground.create_worker()
        # create_worker creates an initial token; since we want to carefully
        # exercise the behaviour of create_worker_token in this test, we
        # need to delete that initial token.
        initial_token = worker.token
        assert initial_token is not None
        token = playground.create_worker_token(worker=worker)
        initial_token.delete()
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertEqual(token.worker, worker)
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_worker_token_disabled(self) -> None:
        """create_worker_token returns a disabled token."""
        playground = Playground()
        token = playground.create_worker_token(enabled=False)
        self.assertFalse(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertIsNotNone(token.worker)
        self.assertFalse(hasattr(token, "activating_worker"))
        self.assertFalse(token.explicit_grants_only)

    def test_create_worker_activation_token(self) -> None:
        """create_worker_activation_token returns an enabled token."""
        playground = Playground()
        token = playground.create_worker_activation_token()
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER_ACTIVATION)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertIsNotNone(token.activating_worker)
        self.assertFalse(token.explicit_grants_only)

    def test_create_worker_activation_token_with_worker(self) -> None:
        """create_worker_activation_token returns a token with a worker."""
        playground = Playground()
        worker = playground.create_worker()
        token = playground.create_worker_activation_token(worker=worker)
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER_ACTIVATION)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertEqual(token.activating_worker, worker)
        self.assertFalse(token.explicit_grants_only)

    def test_create_worker_activation_token_disabled(self) -> None:
        """create_worker_activation_token returns a disabled token."""
        playground = Playground()
        token = playground.create_worker_activation_token(enabled=False)
        self.assertFalse(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORKER_ACTIVATION)
        self.assertIsNone(token.user)
        self.assertEqual(token.external_username, "")
        self.assertFalse(hasattr(token, "worker"))
        self.assertIsNotNone(token.activating_worker)
        self.assertFalse(token.explicit_grants_only)

    def test_create_work_request_token(self) -> None:
        playground = Playground()
        token = playground.create_work_request_token()
        self.assertTrue(token.enabled)
        self.assertEqual(token.token_type, Token.Types.WORK_REQUEST)
        self.assertIsNone(token.user)
        self.assertFalse(hasattr(token, "worker"))
        assert token.work_request
        self.assertEqual(token, token.work_request.work_request_token)

    def test_create_work_request_token_with_work_request(self) -> None:
        playground = Playground()
        work_request = playground.create_worker_task(
            mark_pending=True, assign_new_worker=True
        )
        token = playground.create_work_request_token(work_request)
        self.assertEqual(token.work_request, work_request)

    def test_create_work_request_token_with_worker(self) -> None:
        playground = Playground()
        worker = playground.create_worker()
        token = playground.create_work_request_token(worker=worker)
        self.assertTrue(token.enabled)
        assert token.work_request
        self.assertEqual(token.work_request.worker, worker)

    def test_create_group(self) -> None:
        playground = Playground()
        scope = playground.get_default_scope()
        user1 = playground.create_user("user1")
        group = playground.create_group("testgroup", users=[user1], scope=scope)
        self.assertEqual(group.name, "testgroup")
        self.assertQuerySetEqual(group.users.all(), [user1])
        self.assertEqual(group.scope, scope)

    def test_create_group_defaults(self) -> None:
        playground = Playground()
        group = playground.create_group("testgroup")
        self.assertEqual(group.name, "testgroup")
        self.assertQuerySetEqual(group.users.all(), [])
        self.assertEqual(group.scope, playground.get_default_scope())

    def test_create_group_role(self) -> None:
        """Test create_group_role."""
        playground = Playground()
        scope = playground.get_default_scope()
        group = playground.create_group_role(scope, Scope.Roles.OWNER)
        self.assertEqual(group.name, f"{scope.name}-owner")
        self.assertQuerySetEqual(group.users.all(), [])
        assignment = ScopeRole.objects.get(resource=scope, group=group)
        self.assertEqual(assignment.role, Scope.Roles.OWNER)

    def test_create_group_role_with_name(self) -> None:
        """Test create_group_role with name."""
        playground = Playground()
        scope = playground.get_default_scope()
        group = playground.create_group_role(
            scope, Scope.Roles.OWNER, name="foo"
        )
        self.assertEqual(group.name, "foo")
        self.assertQuerySetEqual(group.users.all(), [])
        assignment = ScopeRole.objects.get(resource=scope, group=group)
        self.assertEqual(assignment.role, Scope.Roles.OWNER)

    def test_create_group_role_with_users(self) -> None:
        """Test create_group_role with name."""
        playground = Playground()
        scope = playground.get_default_scope()
        user1 = playground.get_default_user()
        user2 = playground.create_user("user2")
        group = playground.create_group_role(
            scope, Scope.Roles.OWNER, users=[user1, user2]
        )
        self.assertQuerySetEqual(
            group.users.all(), [user1, user2], ordered=False
        )
        assignment = ScopeRole.objects.get(resource=scope, group=group)
        self.assertEqual(assignment.role, Scope.Roles.OWNER)

    def test_create_group_role_idempotent(self) -> None:
        """Test create_group_role being idempotent."""
        playground = Playground()
        scope = playground.get_default_scope()
        user1 = playground.get_default_user()
        user2 = playground.create_user("user2")

        group = playground.create_group_role(
            scope, Scope.Roles.OWNER, users=[user1]
        )
        assignment = ScopeRole.objects.get(resource=scope, group=group)
        self.assertEqual(assignment.role, Scope.Roles.OWNER)
        self.assertQuerySetEqual(group.users.all(), [user1])

        group2 = playground.create_group_role(
            scope, Scope.Roles.OWNER, users=[user2]
        )
        self.assertEqual(group.pk, group2.pk)
        self.assertQuerySetEqual(group2.users.all(), [user2])

        assignment2 = ScopeRole.objects.get(resource=scope, group=group)
        # Assignment is regenerated
        self.assertNotEqual(assignment.pk, assignment2.pk)
        self.assertEqual(assignment.role, Scope.Roles.OWNER)

    def test_scenario(self) -> None:
        """Test building a scenario."""
        playground = Playground()
        scenario = scenarios.DefaultContext()
        playground.build_scenario(scenario)
        self.assertIsNotNone(scenario.scope)
        self.assertEqual(playground.scenarios, {})
        self.assertIsNone(context.scope)

    def test_scenario_named(self) -> None:
        """Test building a named scenario."""
        playground = Playground()
        scenario = scenarios.DefaultContext()
        playground.build_scenario(scenario, scenario_name="test")
        self.assertIsNotNone(scenario.scope)
        self.assertEqual(playground.scenarios, {"test": scenario})
        self.assertIsNone(context.scope)

    def test_scenario_set_current(self) -> None:
        """Test building a scenario with set_current."""
        with context.local():
            playground = Playground()
            scenario = scenarios.DefaultContext(set_current=True)
            playground.build_scenario(
                scenario, scenario_name="test", set_current=True
            )
            self.assertEqual(playground.scenarios, {"test": scenario})
            self.assertEqual(context.scope, scenario.scope)

    def test_scenario_set_current_not_requested(self) -> None:
        """Test building a scenario with set_current."""
        with context.local():
            playground = Playground()
            scenario = scenarios.DefaultContext()
            playground.build_scenario(
                scenario, scenario_name="test", set_current=True
            )
            self.assertEqual(playground.scenarios, {"test": scenario})
            self.assertIsNone(context.scope)

    def test_scenario_set_current_not_yet(self) -> None:
        """Test building a scenario with set_current."""
        with context.local():
            playground = Playground()
            scenario = scenarios.DefaultContext(set_current=True)
            playground.build_scenario(scenario, scenario_name="test")
            self.assertEqual(playground.scenarios, {"test": scenario})
            self.assertIsNone(context.scope)

    def test_create_worker_for_work_request_invalid_type(self) -> None:
        playground = Playground()
        package = playground.create_upload_artifacts().upload
        for wr in (
            playground.create_internal_task(task_name="workflow"),
            playground.create_wait_task(
                task_name="externaldebsign",
                task_data=ExternalDebsignData(unsigned=package.pk),
                workflow_data=WorkRequestWorkflowData(needs_input=True),
            ),
            playground.create_workflow(),
        ):
            with (
                self.subTest(task_type=str(wr.task_type)),
                self.assertRaisesRegex(
                    NotImplementedError,
                    "Cannot create a worker that can run"
                    f" a {wr.task_type} work request",
                ),
            ):
                playground._create_worker_for_work_request(wr)

    def test_schedule_for_worker_success(self) -> None:
        playground = Playground()
        worker = playground.create_worker()
        work_request = playground.create_worker_task(mark_pending=True)

        self.enterContext(
            mock.patch(
                "debusine.server.scheduler.get_candidate_work_requests",
                return_value=WorkRequest.objects.filter(pk=work_request.pk),
            )
        )
        self.enterContext(
            mock.patch(
                "debusine.db.tasks.DBProxyTask.can_run_on",
                return_value=True,
            )
        )
        self.assertEqual(playground.schedule_for_worker(worker), work_request)

    def test_schedule_for_worker_task_error(self) -> None:
        playground = Playground()
        worker = playground.create_worker()
        work_request = playground.create_worker_task(mark_pending=True)

        self.enterContext(
            mock.patch(
                "debusine.server.scheduler.get_candidate_work_requests",
                return_value=WorkRequest.objects.filter(pk=work_request.pk),
            )
        )
        get_task = self.enterContext(
            mock.patch(
                "debusine.db.models.WorkRequest.get_task",
                side_effect=TaskConfigError("expected failure"),
            )
        )
        with self.assertRaisesRegex(AssignToWorkerError, "expected failure"):
            playground.schedule_for_worker(worker)
        get_task.assert_called()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
A fake SFTP server for testing.

Do not use this in production!  It has only the features required for the
tests that currently use it, and no security.
"""

import io
import os
import socket
import time as tm
from threading import Event, Thread
from typing import Self, override

from cryptography.hazmat.primitives import asymmetric, serialization
from paramiko import (
    Ed25519Key,
    PKey,
    SFTPAttributes,
    SFTPHandle,
    SFTPServer,
    SFTPServerInterface,
    ServerInterface,
    Transport,
)
from paramiko.common import (
    AUTH_FAILED,
    AUTH_SUCCESSFUL,
    OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED,
    OPEN_SUCCEEDED,
)
from paramiko.sftp import SFTP_FAILURE, SFTP_OK


class FakeSSHServer(ServerInterface):
    """A fake SSH server for testing."""

    def __init__(self, user_key: Ed25519Key) -> None:
        """Construct the server interface with a single permitted user key."""
        self.user_key = user_key

    @override
    def check_channel_request(self, kind: str, chanid: int) -> int:
        """Check if a channel request is granted."""
        chanid  # fake usage for vulture
        return (
            OPEN_SUCCEEDED
            if kind in {"session", "subsystem"}
            else OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED
        )

    @override
    def get_allowed_auths(
        self,
        username: str,  # noqa: ARG002, U100
    ) -> str:  # pragma: no cover
        """Return supported authentication methods."""
        return "publickey"

    @override
    def check_auth_publickey(
        self,
        username: str,  # noqa: ARG002, U100
        key: PKey,
    ) -> int:
        """Determine if a client key is acceptable for authentication."""
        return AUTH_SUCCESSFUL if key == self.user_key else AUTH_FAILED


class FakeSFTPServer(SFTPServerInterface):
    """A fake SFTP subsystem for testing."""

    @override
    def open(
        self, path: str, flags: int, attr: SFTPAttributes | None
    ) -> SFTPHandle | int:
        """Open a file on the server and create a handle."""
        assert flags & os.O_WRONLY
        assert flags & os.O_CREAT
        assert attr is not None
        SFTPServer.set_file_attr(path, attr)
        # This will be closed by SFTPHandle.close.
        f = open(path, "wb")  # noqa: SIM115
        handle = SFTPHandle(flags)
        handle.writefile = f  # type: ignore[attr-defined]
        return handle

    @override
    def stat(self, path: str) -> SFTPAttributes | int:
        """Return attributes of a path on the server."""
        try:
            return SFTPAttributes.from_stat(os.stat(path))
        except OSError as e:
            return (
                SFTPServer.convert_errno(e.errno)
                if e.errno is not None
                else SFTP_FAILURE
            )

    @override
    def chattr(self, path: str, attr: SFTPAttributes) -> int:
        """Change the attributes of a file."""
        SFTPServer.set_file_attr(path, attr)
        return SFTP_OK


class WrappedEd25519Key:
    """A wrapper for the various formats of Ed25519 key that we need."""

    def __init__(
        self, private_key: asymmetric.ed25519.Ed25519PrivateKey
    ) -> None:
        """Construct a wrapped key object."""
        self.private_key = private_key

    @classmethod
    def generate(cls) -> Self:
        """Generate an Ed25519 private key."""
        return cls(asymmetric.ed25519.Ed25519PrivateKey.generate())

    @property
    def private_openssh(self) -> str:
        """Return the OpenSSH private key representation of this key."""
        return self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.OpenSSH,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode()

    @property
    def public_openssh(self) -> str:  # pragma: no cover
        """Return the OpenSSH public key representation of this key."""
        return (
            self.private_key.public_key()
            .public_bytes(
                encoding=serialization.Encoding.OpenSSH,
                format=serialization.PublicFormat.OpenSSH,
            )
            .decode()
        )

    @property
    def public_paramiko(self) -> Ed25519Key:
        """Return the Paramiko public key representation of this key."""
        return Ed25519Key.from_private_key(io.StringIO(self.private_openssh))


class FakeSFTPServerThread(Thread):
    """Thread that runs a fake SFTP server for testing."""

    def __init__(self) -> None:
        """Initialize the server and start listening."""
        super().__init__()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(("", 0))
        self.socket.listen()
        self.host_key = WrappedEd25519Key.generate()
        self.user_key = WrappedEd25519Key.generate()
        self.stop_event = Event()

    @property
    def address(self) -> tuple[str, int]:
        """Return the (host, port) to be used by clients."""
        host, port = self.socket.getsockname()
        assert isinstance(host, str)
        assert isinstance(port, int)
        return host, port

    @override
    def run(self) -> None:
        """Run the server."""
        while not self.stop_event.is_set():
            self.socket.settimeout(1)
            try:
                conn, _ = self.socket.accept()
            except TimeoutError:  # pragma: no cover
                continue
            self.socket.settimeout(None)
            transport = Transport(conn)
            transport.add_server_key(self.host_key.public_paramiko)
            transport.set_subsystem_handler("sftp", SFTPServer, FakeSFTPServer)
            server = FakeSSHServer(self.user_key.public_paramiko)
            transport.start_server(server=server)
            channel = transport.accept()
            if channel is not None:  # pragma: no cover
                try:
                    while transport.is_active():
                        tm.sleep(0.1)
                finally:
                    transport.close()
                    channel.close()

    def stop(self) -> None:
        """Stop the server."""
        self.stop_event.set()
        self.join()
        self.socket.close()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the artifact views."""

import abc
import datetime as dt
import io
import json
import os.path
import tarfile
from contextlib import ExitStack
from pathlib import Path
from typing import Any, ClassVar, cast, override
from unittest import mock

import lxml
from django.db.models import Max, TextChoices
from django.db.models.functions import Lower
from django.http.response import HttpResponseBase
from django.test import Client, RequestFactory
from django.urls import reverse
from django.utils.http import http_date
from django.utils.timesince import timesince
from rest_framework import status

from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    FileInArtifact,
    Token,
    User,
    WorkRequest,
    Workspace,
)
from debusine.server.tar import TarArtifact
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)
from debusine.web.views.artifacts import (
    ArtifactDetailView,
    FileDetailView,
    FileDetailViewRaw,
)
from debusine.web.views.tests.utils import ViewTestMixin, date_format


class PermissionTests(TestCase, abc.ABC):
    """Permission checks common to all other test cases."""

    # Note: this is deleted at the end of the file, to prevent it from being
    # run as a test case

    user: ClassVar[User]

    @abc.abstractmethod
    def permission_tests_get(
        self, *, include_token: bool, public_workspace: bool = False
    ) -> TestResponseType:
        """Override to perform a get request to drive permission tests."""

    def test_check_denied_private_workspace(self) -> None:
        """Permission denied: no token, logged user or public workspace."""
        response = self.permission_tests_get(include_token=False)
        self.assertContains(
            response,
            "Workspace System not found in scope debusine, or you are not "
            "authorized to see it",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_check_permissions_valid_token_allowed(self) -> None:
        """Permission granted: valid token."""
        response = self.permission_tests_get(include_token=True)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_check_permissions_public_workspace(self) -> None:
        """Permission granted: without a token but it is a public workspace."""
        response = self.permission_tests_get(
            include_token=False, public_workspace=True
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_check_permission_logged_user(self) -> None:
        """Permission granted: user is logged in."""
        self.client.force_login(self.user)

        response = self.permission_tests_get(include_token=False)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_check_can_display(self) -> None:
        """Check that Artifact.can_display is honored."""
        with override_permission(Artifact, "can_display", AllowAll):
            response = self.permission_tests_get(
                include_token=False, public_workspace=True
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        with override_permission(Artifact, "can_display", DenyAll):
            response = self.permission_tests_get(
                include_token=False, public_workspace=True
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class ArtifactDetailViewTests(ViewTestMixin, PermissionTests):
    """Tests for the ArtifactDetailView class."""

    user: ClassVar[User]
    token: ClassVar[Token]
    path_in_artifact: ClassVar[str]
    file_size: ClassVar[int]
    artifact: ClassVar[Artifact]
    files_contents: ClassVar[dict[str, bytes]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.token = cls.playground.create_user_token(user=cls.user)
        cls.path_in_artifact = "README.md"
        cls.file_size = 100
        cls.artifact, cls.files_contents = cls.playground.create_artifact(
            [cls.path_in_artifact, "AUTHORS"],
            files_size=cls.file_size,
            expiration_delay=1,
            create_files=True,
        )
        cls.playground.create_group_role(
            cls.artifact.workspace, Workspace.Roles.OWNER, users=[cls.user]
        )

    @override
    def permission_tests_get(
        self, *, include_token: bool, public_workspace: bool = False
    ) -> TestResponseType:
        """Perform a get request to drive permission tests."""
        if public_workspace:
            self.assertTrue(self.artifact.workspace.public)
        else:
            self.artifact.workspace.public = False
            self.artifact.workspace.save()

        headers: dict[str, Any] = {}
        if include_token:
            headers["HTTP_TOKEN"] = self.token.key
        return self.client.get(
            self.artifact.get_absolute_url(),
            **headers,
        )

    def _get(
        self,
        pk: int | None = None,
    ) -> TestResponseType:
        """GET request on the ArtifactDetail view."""
        if pk is None:
            pk = self.artifact.pk
        return self.client.get(
            reverse(
                "workspaces:artifacts:detail",
                kwargs={
                    "wname": self.playground.get_default_workspace().name,
                    "artifact_id": pk,
                },
            ),
            headers={"token": self.token.key},
        )

    def assertFile(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        artifact: Artifact,
        *,
        suffix: str | None = None,
    ) -> None:
        """Ensure that the single file in the artifact is shown."""
        file = artifact.fileinartifact_set.select_related("file").get()
        div_id = (
            "formatted-content"
            if suffix is None
            else f"formatted-content-{suffix}"
        )
        formatted_content = self.assertHasElement(
            tree, f"//div[@id='{div_id}']"
        )

        self.assertTextContentEqual(formatted_content.div[0].div[0], file.path)
        gadgets = formatted_content.div[0].div[1]

        if file.complete:
            self.assertEqual(gadgets.get("class"), "btn-group btn-group-sm")
            self.assertEqual(
                gadgets.a[0].get("href"), file.get_absolute_url_raw()
            )
            self.assertEqual(
                gadgets.a[1].get("href"), file.get_absolute_url_download()
            )
            card_body_class = formatted_content.div[1].get("class")
            assert card_body_class is not None
            self.assertIn("card-body", card_body_class.split())
        else:
            self.assertTextContentEqual(gadgets, "incomplete")
            self.assertEqual(len(formatted_content.div), 1)

    def assertFileList(
        self, tree: lxml.objectify.ObjectifiedElement, artifact: Artifact
    ) -> None:
        """Ensure there is a file list with all files in the artifact."""
        files = artifact.fileinartifact_set.select_related("file").order_by(
            Lower("path")
        )
        table = self.assertHasElement(tree, "//table[@id='file-list']")

        for tr, file in zip(table.tbody.tr, files):
            with self.subTest(file.path):
                basename = os.path.basename(file.path)
                if file.complete:
                    self.assertEqual(
                        tr.td[0].a.get("href"), file.get_absolute_url()
                    )
                    self.assertTextContentEqual(tr.td[0].a, basename)
                else:
                    self.assertEqual(tr.td[0].countchildren(), 0)
                    self.assertTextContentEqual(
                        tr.td[0], f"{basename} (incomplete)"
                    )
                self.assertEqual(tr.td[1].get("title"), str(file.file.size))

    def assertRelations(
        self, tree: lxml.objectify.ObjectifiedElement, artifact: Artifact
    ) -> None:
        """Ensure that there is a relation list with all the relations."""
        relations = artifact.relations.all()
        targeted_by = artifact.targeted_by.filter(
            type=ArtifactRelation.Relations.EXTENDS
        )
        num_relations = len(relations) + len(targeted_by)
        tables = tree.xpath("//table[@id='relation-list']")

        relations_a = self.assertHasElement(
            tree, "//a[@id='nav-relations-tab']"
        )

        if num_relations == 1:
            self.assertTextContentEqual(relations_a, "Relation (1)")
        else:
            self.assertTextContentEqual(
                relations_a, f"Relations ({num_relations})"
            )

        button_classes = relations_a.get("class")
        assert button_classes is not None

        if num_relations == 0:
            self.assertIn("disabled", button_classes.split())

            # No "relation-list" tables
            self.assertFalse(tables)

            div = tree.xpath("//div[@id='nav-relations']")
            self.assertTextContentEqual(div[0], "No relations.")
            return

        self.assertNotIn("disabled", button_classes.split())

        table = tables[0]
        self.assertEqual(len(table.tbody.tr), num_relations)
        for tr, relation in zip(table.tbody.tr, relations):
            with self.subTest(str(tr)):
                self.assertEqual(tr.td[0].text, relation.type)
                self.assertEqual(
                    tr.td[1].i.attrib["class"], "bi bi-arrow-right"
                )
                label = relation.target.get_label()
                self.assertEqual(tr.td[3].a.text, label)
                if relation.target.fileinartifact_set.filter(
                    complete=False
                ).exists():
                    self.assertTextContentEqual(
                        tr.td[3], f"{label} (incomplete)"
                    )
                else:
                    self.assertTextContentEqual(tr.td[3], label)
        for tr, relation in zip(table.tbody.tr[len(relations) :], targeted_by):
            with self.subTest(str(tr)):
                self.assertEqual(tr.td[0].text, relation.type)
                self.assertEqual(tr.td[1].i.attrib["class"], "bi bi-arrow-left")
                label = relation.artifact.get_label()
                self.assertEqual(tr.td[3].a.text, label)
                if relation.artifact.fileinartifact_set.filter(
                    complete=False
                ).exists():
                    self.assertTextContentEqual(
                        tr.td[3], f"{label} (incomplete)"
                    )
                else:
                    self.assertTextContentEqual(tr.td[3], label)

    def test_permissions(self) -> None:
        """Test basic permission enforcement."""
        self.assertSetsCurrentWorkspace(
            self.artifact.workspace, self.artifact.get_absolute_url()
        )
        self.assertEnforcesPermission(
            self.artifact.can_display,
            self.artifact.get_absolute_url(),
            "get_context_data",
        )

    def test_get_place(self) -> None:
        view = ArtifactDetailView()
        view.object = self.artifact
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, "Artifact debusine:test")
        self.assertEqual(place.url, self.artifact.get_absolute_url())
        self.assertEqual(place.icon, "file")
        self.assertEqual(
            place.description, "debusine:test artifact debusine:test"
        )
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.artifact.workspace.get_absolute_url()
        )

    def test_invalid_artifact_id(self) -> None:
        """Test viewing an artifact ID that does not exist."""
        artifact_id = Artifact.objects.aggregate(Max("id"))['id__max'] + 1
        response = self._get(pk=artifact_id)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"],
            "No artifact found matching the query",
        )

    def test_get_success_html_files_list(self) -> None:
        """View shows a list of files."""
        work_request = self.playground.create_worker_task(task_name="noop")
        with context.disable_permission_checks():
            self.artifact.created_by_work_request = work_request
            self.artifact.save()
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertFileList(tree, self.artifact)
        self.assertRelations(tree, self.artifact)
        self.assertFalse(tree.xpath("//div[@id='formatted-content']"))

        files_a = self.assertHasElement(tree, "//a[@id='nav-files-tab']")
        assert files_a is not None
        self.assertTextContentEqual(files_a, "Files (2)")

        # Assert file list is displayed by default
        file_list = self.assertHasElement(tree, "//div[@id='nav-files']")
        self.assertIn("show active", file_list.attrib["class"])

    def test_get_success_html_files_list_incomplete(self) -> None:
        """View shows incomplete files in a list differently."""
        self.artifact.fileinartifact_set.filter(path="AUTHORS").update(
            complete=False
        )
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertFileList(tree, self.artifact)
        self.assertRelations(tree, self.artifact)
        self.assertFalse(tree.xpath("//div[@id='formatted-content']"))

    def test_get_success_html_singlefile(self) -> None:
        """View show the content of the only file in the artifact."""
        self.artifact.fileinartifact_set.filter(path="AUTHORS").delete()
        response = self._get()
        tree = self.assertResponseHTML(response)
        files_a = self.assertHasElement(tree, "//a[@id='nav-files-tab']")
        assert files_a is not None
        self.assertTextContentEqual(files_a, "File (1)")

        file_classes = files_a.get("class")
        assert file_classes is not None
        self.assertNotIn("disabled", file_classes.split())
        self.assertFalse(tree.xpath("//table[@id='file-list']"))
        self.assertFile(tree, self.artifact, suffix="files")

    def test_get_success_html_singlefile_incomplete(self) -> None:
        """View shows a single incomplete file in the artifact differently."""
        self.artifact.fileinartifact_set.filter(path="AUTHORS").delete()
        self.artifact.fileinartifact_set.update(complete=False)
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='file-list']"))
        self.assertFile(tree, self.artifact, suffix="files")

    def test_get_success_html_empty_artifact(self) -> None:
        """Test HTML output if there are no files in the artifact."""
        artifact, _ = self.playground.create_artifact([])
        response = self._get(artifact.id)
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='file-list']"))
        self.assertFalse(tree.xpath("//div[@id='formatted-content']"))

        files_a = self.assertHasElement(tree, "//a[@id='nav-files-tab']")
        self.assertTextContentEqual(files_a, "Files (0)")
        files_classes = files_a.get("class")
        assert files_classes is not None
        self.assertIn("disabled", files_classes.split())

        div = tree.xpath("//div[@id='nav-files']")
        self.assertTextContentEqual(div[0], "No files.")

    def test_get_success_html_user_and_no_expiration(self) -> None:
        """Test HTML output with no user and expiration."""
        with context.disable_permission_checks():
            self.artifact.created_by = self.user
            self.artifact.expiration_delay = dt.timedelta(0)
            self.artifact.save()
        response = self._get()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_success_html_fwd_relations(self) -> None:
        """View shows a list of forward relations."""
        other_artifact = self.playground.create_build_log_artifact()
        self.playground.create_artifact_relation(self.artifact, other_artifact)
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertRelations(tree, self.artifact)

    def test_get_success_html_fwd_relations_incomplete(self) -> None:
        """View shows whether forward-related artifacts are incomplete."""
        other_artifact = self.playground.create_build_log_artifact(
            skip_add_files_in_store=True
        )
        self.playground.create_artifact_relation(self.artifact, other_artifact)
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertRelations(tree, self.artifact)

    def test_get_success_html_rev_relations(self) -> None:
        """View shows a list of reverse relations."""
        other_artifact = self.playground.create_build_log_artifact()
        self.playground.create_artifact_relation(other_artifact, self.artifact)
        self.playground.create_artifact_relation(
            other_artifact,
            self.artifact,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertRelations(tree, self.artifact)

    def test_get_success_html_rev_relations_incomplete(self) -> None:
        """View shows whether reverse-related artifacts are incomplete."""
        other_artifact = self.playground.create_build_log_artifact(
            skip_add_files_in_store=True
        )
        self.playground.create_artifact_relation(other_artifact, self.artifact)
        self.playground.create_artifact_relation(
            other_artifact,
            self.artifact,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )
        response = self._get()
        tree = self.assertResponseHTML(response)
        self.assertRelations(tree, self.artifact)

    def test_metadata(self) -> None:
        tree = self.assertResponseHTML(self._get())
        table = self.assertHasElement(tree, "//table[@id='metadata-table']")

        artifact = self.artifact

        # Assert Category
        category = self.assertHasElement(table, ".//tr[1]/td[1]")
        self.assertTextContentEqual(category, artifact.category)

        # Assert Workspace
        workspace_link = self.assertHasElement(table, ".//tr[2]/td[1]/a")
        self.assertEqual(
            workspace_link.get("href"), artifact.workspace.get_absolute_url()
        )

        self.assertTextContentEqual(workspace_link, str(artifact.workspace))

        # Assert Created at
        created_at = self.assertHasElement(table, ".//tr[3]/td[1]")
        self.assertTextContentEqual(
            created_at, f"{timesince(artifact.created_at)} ago"
        )

        # Assert user
        user = self.assertHasElement(table, ".//tr[4]/td[1]")
        self.assertTextContentEqual(user, str(artifact.created_by))

        # Assert Expiration date
        expiration = self.assertHasElement(table, ".//tr[5]/td[1]")
        assert artifact.expire_at is not None
        self.assertTextContentEqual(expiration, date_format(artifact.expire_at))

        # Assert Data
        data = self.assertHasElement(table, ".//tr[7]/td[1]")
        self.assertEqual(
            json.loads(self.get_node_text_normalized(data)), artifact.data
        )

    def test_origin_created_by_user(self) -> None:
        self.artifact.created_by = self.user
        self.artifact.save()

        tree = self.assertResponseHTML(self._get())

        created_by = self.assertHasElement(
            tree, "//p[@id='artifact-created-by-user']"
        )
        self.assertTextContentEqual(
            created_by, f"Artifact created by {self.user}."
        )
        self.assertEqual(
            created_by.a.attrib["href"], self.user.get_absolute_url()
        )

    def test_origin_work_request_no_workflow(self) -> None:
        work_request = self.playground.create_worker_task(
            task_name="noop", assign_new_worker=True
        )
        with context.disable_permission_checks():
            self.artifact.created_by_work_request = work_request
            self.artifact.save()

        tree = self.assertResponseHTML(self._get())
        table = self.assertHasElement(tree, "//table[@id='work_request-table']")

        # Assert work_request link
        work_request_link = self.assertHasElement(table, ".//tr[1]/td[1]/a")
        self.assertEqual(
            work_request_link.get("href"), work_request.get_absolute_url()
        )
        self.assertTextContentEqual(
            work_request_link, f"{work_request.id} ({work_request.task_name})"
        )

        # Assert Status and result
        status_result = self.assertHasElement(table, ".//tr[2]/td[1]")
        work_request_status = WorkRequest.Statuses(work_request.status)
        work_request_result = WorkRequest.Results(work_request.result)
        self.assertTextContentEqual(
            status_result,
            f"{work_request_status.label} {work_request_result.label}",
        )

        # Assert workspace link
        workspace_link = self.assertHasElement(table, ".//tr[3]/td[1]/a")
        self.assertEqual(
            workspace_link.get("href"),
            work_request.workspace.get_absolute_url(),
        )
        self.assertTextContentEqual(workspace_link, str(work_request.workspace))

        # Assert created by user
        user = self.assertHasElement(table, ".//tr[4]/td[1]")
        self.assertTextContentEqual(user, str(work_request.created_by))

        # Assert created at
        created_at = self.assertHasElement(table, ".//tr[5]/td[1]")
        self.assertTextContentEqual(
            created_at, f"{timesince(work_request.created_at)} ago"
        )

        # Assert worker
        worker = self.assertHasElement(table, ".//tr[6]/td[1]")
        assert work_request.worker
        self.assertTextContentEqual(worker, work_request.worker.name)

        # Assert no workflow table
        with self.assertRaises(AssertionError):
            self.assertHasElement(tree, "//table[@id='workflow-table']")

    def test_origin_work_request_workflow(self) -> None:
        workflow = self.playground.create_workflow()

        work_request = self.playground.create_worker_task(
            task_name="noop", assign_new_worker=True, parent=workflow
        )
        with context.disable_permission_checks():
            self.artifact.created_by_work_request = work_request
            self.artifact.save()

        tree = self.assertResponseHTML(self._get())

        # Has the work_request-table, tested in
        # test_origin_work_request_no_workflow()
        self.assertHasElement(tree, "//table[@id='work_request-table']")

        table = self.assertHasElement(tree, "//table[@id='workflow-table']")

        # Assert workflow link
        workflow_link = self.assertHasElement(table, ".//tr[1]/td[1]/a")
        self.assertEqual(workflow_link.get("href"), workflow.get_absolute_url())
        self.assertTextContentEqual(
            workflow_link, f"{workflow.id} ({workflow.task_name})"
        )

        # Assert status and result
        status_result = self.assertHasElement(table, ".//tr[2]/td[1]")
        workflow_status = cast(TextChoices, workflow.status)
        workflow_result = cast(TextChoices, workflow.result)
        self.assertTextContentEqual(
            status_result,
            f"{workflow_status.label} {workflow_result.label}",
        )

        # Assert workspace link
        workspace_link = self.assertHasElement(table, ".//tr[3]/td[1]/a")
        self.assertEqual(
            workspace_link.get("href"),
            workflow.workspace.get_absolute_url(),
        )
        self.assertTextContentEqual(workspace_link, str(workflow.workspace))

        # Assert created by user
        user = self.assertHasElement(table, ".//tr[4]/td[1]")
        self.assertTextContentEqual(user, str(workflow.created_by))

        # Assert created at
        created_at = self.assertHasElement(table, ".//tr[5]/td[1]")
        self.assertTextContentEqual(
            created_at, f"{timesince(workflow.created_at)} ago"
        )


class FileDetailViewTests(ViewTestMixin, PermissionTests):
    """Test FileDetailView."""

    user: ClassVar[User]
    token: ClassVar[Token]
    contents: ClassVar[dict[str, bytes]]
    artifact: ClassVar[Artifact]
    file: ClassVar[FileInArtifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.token = cls.playground.create_user_token(user=cls.user)
        cls.contents = {
            "file.md": b"Line 1\nLine 2\n",
        }
        cls.artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.file = cls.artifact.fileinartifact_set.get(path="file.md")
        cls.playground.create_group_role(
            cls.artifact.workspace, Workspace.Roles.OWNER, users=[cls.user]
        )

    @override
    def permission_tests_get(
        self, *, include_token: bool, public_workspace: bool = False
    ) -> TestResponseType:
        """Perform a get request to drive permission tests."""
        if public_workspace:
            self.assertTrue(self.artifact.workspace.public)
        else:
            self.artifact.workspace.public = False
            self.artifact.workspace.save()

        headers: dict[str, Any] = {}
        if include_token:
            headers["HTTP_TOKEN"] = self.token.key
        return self.client.get(
            self.file.get_absolute_url(),
            **headers,
        )

    def test_permissions(self) -> None:
        """Test basic permission enforcement."""
        url = self.file.get_absolute_url()
        self.assertSetsCurrentWorkspace(self.artifact.workspace, url)
        self.assertEnforcesPermission(
            self.artifact.can_display,
            url,
            "get_context_data",
        )

    def test_get_place(self) -> None:
        view = FileDetailView()
        view.object = self.file
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.file.path)
        self.assertEqual(place.url, self.file.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_get_place_incomplete_file(self) -> None:
        self.file.complete = False
        view = FileDetailView()
        view.object = self.file
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.file.path + " (incomplete)")
        self.assertEqual(place.url, self.file.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_invalid_artifact_id(self) -> None:
        """Test viewing an artifact ID that does not exist."""
        artifact_id = Artifact.objects.aggregate(Max("id"))['id__max'] + 1
        response = self.client.get(
            reverse(
                "workspaces:artifacts:file-detail",
                kwargs={
                    "wname": self.playground.get_default_workspace().name,
                    "artifact_id": artifact_id,
                    "path": self.file.path,
                },
            ),
            headers={"token": self.token.key},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"],
            "No FileInArtifact matches the given query.",
        )

    def test_invalid_file_in_artifact_path(self) -> None:
        """Test viewing a file_path that does not exist."""
        response = self.client.get(
            reverse(
                "workspaces:artifacts:file-detail",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": "invalid-path",
                },
            ),
            headers={"token": self.token.key},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"],
            "No FileInArtifact matches the given query.",
        )

    def test_get(self) -> None:
        """Test a simple get."""
        response = self.permission_tests_get(include_token=True)
        tree = self.assertResponseHTML(response)
        self.assertIn("file", response.context)
        title = self.assertHasElement(tree, "head/title")
        self.assertTextContentEqual(
            title, f"debusine - System - debusine:test - {self.file.path}"
        )

    def test_get_incomplete(self) -> None:
        """Test a get of an incomplete file."""
        self.file.complete = False
        self.file.save()
        response = self.permission_tests_get(include_token=True)
        tree = self.assertResponseHTML(response)
        self.assertIn("file", response.context)
        title = self.assertHasElement(tree, "head/title")
        self.assertTextContentEqual(
            title,
            "debusine - System - debusine:test"
            f" - {self.file.path} (incomplete)",
        )


class FileDetailViewRawTests(ViewTestMixin, PermissionTests):
    """Test FileDetailRawView."""

    playground_memory_file_store = False

    user: ClassVar[User]
    token: ClassVar[Token]
    contents: ClassVar[dict[str, bytes]]
    artifact: ClassVar[Artifact]
    file: ClassVar[FileInArtifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.token = cls.playground.create_user_token(user=cls.user)
        cls.contents = {
            "file.md": b"Line 1\nLine 2\n",
        }
        cls.artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.file = cls.artifact.fileinartifact_set.get(path="file.md")
        cls.playground.create_group_role(
            cls.artifact.workspace, Workspace.Roles.OWNER, users=[cls.user]
        )

    @override
    def permission_tests_get(
        self, *, include_token: bool, public_workspace: bool = False
    ) -> TestResponseType:
        """Perform a get request to drive permission tests."""
        if public_workspace:
            self.assertTrue(self.artifact.workspace.public)
        else:
            self.artifact.workspace.public = False
            self.artifact.workspace.save()

        headers: dict[str, Any] = {}
        if include_token:
            headers["HTTP_TOKEN"] = self.token.key
        return self.client.get(
            self.file.get_absolute_url_raw(),
            **headers,
        )

    def test_permissions(self) -> None:
        """Test basic permission enforcement."""
        url = self.file.get_absolute_url_raw()
        self.assertSetsCurrentWorkspace(self.artifact.workspace, url)
        self.assertEnforcesPermission(
            self.artifact.can_display,
            url,
            "get",
        )

    def test_get_place(self) -> None:
        view = FileDetailViewRaw()
        view.object = self.file
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.file.path)
        self.assertEqual(place.url, self.file.get_absolute_url_raw())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_get_place_incomplete_file(self) -> None:
        self.file.complete = False
        view = FileDetailViewRaw()
        view.object = self.file
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.file.path + " (incomplete)")
        self.assertEqual(place.url, self.file.get_absolute_url_raw())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_invalid_file_in_artifact_path(self) -> None:
        """Test viewing a file_path that does not exist."""
        response = self.client.get(
            reverse(
                "workspaces:artifacts:file-raw",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": "invalid-path",
                },
            ),
            headers={"token": self.token.key},
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(
            response.context["exception"],
            "No FileInArtifact matches the given query.",
        )

    def test_get(self) -> None:
        """Test a simple get."""
        response = self.permission_tests_get(include_token=True)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Disposition"],
            'inline; filename="file.md"',
        )


class DownloadPathViewTests(ViewTestMixin, PermissionTests):
    """Tests for the DownloadPathView class."""

    playground_memory_file_store = False

    user: ClassVar[User]
    token: ClassVar[Token]
    path_in_artifact: ClassVar[str]
    file_size: ClassVar[int]
    tree_paths: ClassVar[list[str]]
    artifact: ClassVar[Artifact]
    tree: ClassVar[Artifact]
    files_contents: ClassVar[dict[str, bytes]]
    tree_files_contents: ClassVar[dict[str, bytes]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.token = cls.playground.create_user_token(user=cls.user)
        cls.path_in_artifact = "README.md"
        cls.file_size = 100
        cls.artifact, cls.files_contents = cls.playground.create_artifact(
            [cls.path_in_artifact],
            files_size=cls.file_size,
            create_files=True,
        )
        cls.tree_paths = [
            "README",
            "doc/README",
            "doc/README2",
            "documentation",
            "src/lib/main.c",
            "src/lib/utils.c",
        ]
        cls.tree, cls.tree_files_contents = cls.playground.create_artifact(
            cls.tree_paths, create_files=True
        )
        cls.playground.create_group_role(
            cls.artifact.workspace, Workspace.Roles.OWNER, users=[cls.user]
        )

    @override
    def permission_tests_get(
        self, *, include_token: bool, public_workspace: bool = False
    ) -> TestResponseType:
        """Perform a get request to drive permission tests."""
        if public_workspace:
            self.assertTrue(self.artifact.workspace.public)
        else:
            self.artifact.workspace.public = False
            self.artifact.workspace.save()

        headers: dict[str, Any] = {}
        if include_token:
            headers["HTTP_TOKEN"] = self.token.key
        return self.client.get(
            reverse(
                "workspaces:artifacts:file-download",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": "/",
                },
            ),
            **headers,
        )

    def get_file(
        self,
        *,
        artifact_id: int | None = None,
        path_file: str | None = None,
    ) -> HttpResponseBase:
        """
        Download file specified in the parameters.

        Unless specified: try to download the whole file (by default
        self.path_file and self.artifact.id).
        """
        if artifact_id is None:
            artifact_id = self.artifact.id

        if path_file is None:
            path_file = self.path_in_artifact

        return self.client.get(
            reverse(
                "workspaces:artifacts:file-download",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": artifact_id,
                    "path": path_file,
                },
            ),
            headers={"token": self.token.key},
        )

    def get_artifact(
        self,
        artifact: Artifact,
        archive: str | None = None,
        subdirectory: str | None = None,
        **get_kwargs: Any,
    ) -> TestResponseType:
        """Request to download an artifact_id."""
        reverse_kwargs: dict[str, Any] = {
            "wname": artifact.workspace.name,
            "artifact_id": artifact.id,
        }
        viewname = "workspaces:artifacts:download"
        if subdirectory is not None:
            viewname = "workspaces:artifacts:file-download"
            reverse_kwargs["path"] = subdirectory

        if archive is not None:
            get_kwargs["archive"] = archive

        return self.client.get(
            reverse(viewname, kwargs=reverse_kwargs),
            get_kwargs,
            headers={"token": self.token.key},
        )

    def assertFileResponse(
        self, response: HttpResponseBase, status_code: int
    ) -> None:
        """Assert that response has the expected headers and content."""
        self.assertEqual(response.status_code, status_code)
        headers = response.headers

        self.assertEqual(headers["Accept-Ranges"], "bytes")

        file_contents = self.files_contents[self.path_in_artifact]
        response_contents = file_contents

        self.assertEqual(headers["Content-Length"], str(len(response_contents)))
        self.assertEqual(
            headers["Content-Range"],
            f"bytes {0}-{self.file_size - 1}/{self.file_size}",
        )

        filename = Path(self.path_in_artifact).name
        self.assertEqual(
            headers["Content-Disposition"], f'attachment; filename="{filename}"'
        )

        assert hasattr(response, "streaming_content")
        self.assertEqual(
            b"".join(response.streaming_content), response_contents
        )

    def assertResponseDownloadsTree(self, response: HttpResponseBase) -> None:
        """Ensure response is a tar download of the self.tree artifact."""
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        assert hasattr(response, "streaming_content")
        response_content = io.BytesIO(b"".join(response.streaming_content))

        with tarfile.open(fileobj=response_content, mode="r:gz") as tar:
            # Check contents of the tar file
            for path in self.tree_paths:
                reader = tar.extractfile(path)
                assert reader is not None
                self.assertEqual(reader.read(), self.tree_files_contents[path])

        # Check relevant headers
        self.assertEqual(
            response.headers["Content-Type"], "application/octet-stream"
        )
        self.assertEqual(
            response.headers["Content-Disposition"],
            f'attachment; filename="artifact-{self.tree.id}.tar.gz"',
        )
        self.assertEqual(
            response.headers["Last-Modified"],
            http_date(self.tree.created_at.timestamp()),
        )

    def test_permissions(self) -> None:
        """Test basic permission enforcement."""
        url = reverse(
            "workspaces:artifacts:file-download",
            kwargs={
                "wname": self.artifact.workspace.name,
                "artifact_id": self.artifact.id,
                "path": "README",
            },
        )
        self.assertSetsCurrentWorkspace(self.artifact.workspace, url)
        self.assertEnforcesPermission(
            self.artifact.can_display,
            url,
            "get",
        )

    def test_path_url_does_not_end_in_slash(self) -> None:
        """
        URL to download a file does not end in /.

        If ending in / wget or curl -O save the file as index.html
        instead of using Content-Disposition filename.
        """
        url = reverse(
            "workspaces:artifacts:file-download",
            kwargs={"wname": "test", "artifact_id": 10, "path": "package.deb"},
        )
        self.assertFalse(url.endswith("/"))

    def test_get_file(self) -> None:
        """Get return the file."""
        response = self.get_file()
        self.assertFileResponse(response, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_path_artifact_does_not_exist(self) -> None:
        """Get return 404: artifact not found."""
        non_existing_artifact_id = 0

        response = self.get_file(artifact_id=non_existing_artifact_id)

        self.assertContains(
            response,
            f"Artifact {non_existing_artifact_id} does not exist",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_get_file_file_does_not_exist(self) -> None:
        """Get return 404: artifact found but file not found."""
        file_path_no_exist = "does-not-exist"

        response = self.get_file(path_file=file_path_no_exist)

        self.assertContains(
            response,
            f'Artifact {self.artifact.id} does not have '
            f'any file or directory for "{file_path_no_exist}"',
            status_code=status.HTTP_404_NOT_FOUND,
            html=True,
        )

    def test_get_file_incomplete(self) -> None:
        """Get returns 404 for an incomplete file."""
        self.artifact.fileinartifact_set.update(complete=False)

        response = self.get_file()

        self.assertContains(
            response,
            f'Artifact {self.artifact.id} does not have any file or directory '
            f'for "{self.path_in_artifact}"',
            status_code=status.HTTP_404_NOT_FOUND,
            html=True,
        )

    def test_get_subdirectory_does_not_exist_404(self) -> None:
        """View return HTTP 404 Not Found: no files in the subdirectory."""
        subdirectory = "does-not-exist"
        response = self.get_artifact(self.artifact, "tar.gz", subdirectory)

        self.assertContains(
            response,
            f'Artifact {self.artifact.id} does not have any file or '
            f'directory for "{subdirectory}"',
            status_code=status.HTTP_404_NOT_FOUND,
            html=True,
        )

    def test_get_subdirectory_entirely_incomplete(self) -> None:
        """View returns 404 if all files in the subdirectory are incomplete."""
        self.tree.fileinartifact_set.filter(path__startswith="src/lib/").update(
            complete=False
        )

        subdirectory = "src/lib"
        response = self.get_artifact(self.tree, "tar.gz", subdirectory)

        self.assertContains(
            response,
            f'Artifact {self.tree.id} does not have any file or '
            f'directory for "{subdirectory}"',
            status_code=status.HTTP_404_NOT_FOUND,
            html=True,
        )

    def test_get_subdirectory_only_tar_gz(self) -> None:
        """View return tar.gz file with the files from a subdirectory."""
        subdirectory = "src/lib"
        response = self.get_artifact(self.tree, "tar.gz", subdirectory)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Disposition"],
            f'attachment; filename="artifact-{self.tree.id}-src_lib.tar.gz"',
        )
        response_content = io.BytesIO(
            b"".join(getattr(response, "streaming_content"))
        )

        with tarfile.open(fileobj=response_content, mode="r:gz") as tar:
            expected_files = list(
                filter(
                    lambda x: x.startswith(subdirectory + "/"), self.tree_paths
                )
            )
            self.assertEqual(tar.getnames(), expected_files)

    def test_get_subdirectory_only_excludes_incomplete(self) -> None:
        """Downloading a subdirectory excludes incomplete files."""
        excluded_file = self.tree.fileinartifact_set.get(path="src/lib/utils.c")
        excluded_file.complete = False
        excluded_file.save()

        subdirectory = "src/lib"
        response = self.get_artifact(self.tree, "tar.gz", subdirectory)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["Content-Disposition"],
            f'attachment; filename="artifact-{self.tree.id}-src_lib.tar.gz"',
        )
        response_content = io.BytesIO(
            b"".join(getattr(response, "streaming_content"))
        )

        with tarfile.open(fileobj=response_content, mode="r:gz") as tar:
            expected_files = [
                path
                for path in self.tree_paths
                if (
                    path.startswith(subdirectory + "/")
                    and path != excluded_file.path
                )
            ]
            self.assertEqual(tar.getnames(), expected_files)

    def test_get_unsupported_archive_parameter(self) -> None:
        """View return HTTP 400 Bad Request: unsupported archive parameter."""
        archive_format = "tar.xz"
        response = self.get_artifact(self.artifact, archive_format)
        self.assertResponse400(
            response,
            f"Invalid archive parameter: '{archive_format}'. "
            "Supported: auto, tar.gz",
        )

    def test_path_without_archive(self) -> None:
        """Check downloading a path with a missing archive format."""
        response = self.get_artifact(
            self.tree, archive=None, subdirectory="src"
        )
        self.assertResponse400(
            response, "archive argument needed when downloading directories"
        )

    def test_get_artifact_auto_file(self) -> None:
        """Check downloading whole artifact with auto download format."""
        response = self.get_artifact(self.artifact, archive=None)
        self.assertFileResponse(response, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_artifact_auto_only_one_complete_file(self) -> None:
        """Auto download format returns file if only one file is complete."""
        FileInArtifact.objects.create(
            artifact=self.artifact,
            path="incomplete",
            file=self.playground.create_file(b"incomplete"),
            complete=False,
        )
        response = self.get_artifact(self.artifact, archive=None)
        self.assertFileResponse(response, status.HTTP_200_OK)
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_artifact_auto_tree(self) -> None:
        """Check downloading whole artifact with auto download format."""
        response = self.get_artifact(self.tree, archive=None)
        self.assertResponseDownloadsTree(response)

    def test_get_artifact_tar_gz(self) -> None:
        """Download a whole artifact as .tar.gz."""
        response = self.get_artifact(self.tree, "tar.gz")
        self.assertResponseDownloadsTree(response)

    def test_get_artifact_tar_gz_resource_cleanup_on_streaming_interrupted(
        self,
    ) -> None:
        # We create the TarArtifact for the view, to be able to manipulate it
        # later
        tar_artifact = TarArtifact(self.tree)
        self.assertFalse(tar_artifact._closed)
        with mock.patch(
            "debusine.web.views.artifacts.TarArtifact",
            return_value=tar_artifact,
        ) as ta:
            response = self.get_artifact(self.tree, "tar.gz")
        ta.assert_called_with(self.tree, None)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        assert hasattr(response, "streaming_content")

        # Close the response without streaming data
        self.close_streaming_response(response)

        self.assertTrue(tar_artifact._closed)


class DownloadPathViewAuthTests(TestCase):
    """Tests for authorization on the DownloadPathView class."""

    playground_memory_file_store = False

    user: ClassVar[User]
    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.get_default_user()
        cls.artifact, _ = cls.playground.create_artifact(
            ["README"], files_size=100, create_files=True
        )

    def make_private(self) -> None:
        """Make the artifact workspace private."""
        self.artifact.workspace.public = False
        self.artifact.workspace.save()

    def make_owner(self) -> None:
        """Make the user owner of the workspace."""
        self.playground.create_group_role(
            self.artifact.workspace, Workspace.Roles.OWNER, users=[self.user]
        )

    def get(
        self,
        token: Token | None = None,
        user: User | None = None,
        subpath: bool = False,
        mock_input_artifacts: bool = False,
    ) -> bool:
        """Call the whoami API view."""
        client = Client()
        if user is not None:
            client.force_login(user)

        headers = {}
        if token:
            headers["token"] = token.key

        reverse_kwargs: dict[str, Any] = {
            "wname": self.artifact.workspace.name,
            "artifact_id": self.artifact.id,
        }

        if subpath:
            viewname = "workspaces:artifacts:file-download"
            reverse_kwargs["path"] = "README"
        else:
            viewname = "workspaces:artifacts:download"

        with ExitStack() as stack:
            if mock_input_artifacts:
                stack.enter_context(
                    mock.patch(
                        (
                            "debusine.tasks.models.BaseDynamicTaskData"
                            ".get_input_artifacts_ids"
                        ),
                        return_value=[self.artifact.id],
                    )
                )

            response = client.get(
                reverse(viewname, kwargs=reverse_kwargs),
                headers=headers,
            )
        return response.status_code == status.HTTP_200_OK

    def assertAllowed(
        self,
        token: Token | None = None,
        user: User | None = None,
        mock_input_artifacts: bool = False,
    ) -> None:
        """Ensure the given auth credentials pass."""
        for subpath in (False, True):
            self.assertTrue(
                self.get(
                    token=token,
                    user=user,
                    subpath=subpath,
                    mock_input_artifacts=mock_input_artifacts,
                )
            )

    def assertDenied(
        self,
        token: Token | None = None,
        user: User | None = None,
        mock_input_artifacts: bool = False,
    ) -> None:
        """Ensure the given auth credentials do not pass."""
        for subpath in (False, True):
            self.assertFalse(
                self.get(
                    token=token,
                    user=user,
                    subpath=subpath,
                    mock_input_artifacts=mock_input_artifacts,
                )
            )

    def test_worker_token(self) -> None:
        """Try a worker token on a public workspace."""
        token = self.playground.create_worker_token()
        self.assertDenied(token)

    def test_public_with_work_request_token(self) -> None:
        """Try a work request token on a public workspace."""
        token = self.playground.create_work_request_token()
        self.assertAllowed(token, mock_input_artifacts=True)

    def test_private_with_work_request_token(self) -> None:
        """Try a request token on a private workspace."""
        token = self.playground.create_work_request_token()
        self.playground.create_group_role(
            self.artifact.workspace, Workspace.Roles.VIEWER, users=[self.user]
        )
        self.make_private()
        with self.subTest(mock_input_artifacts=False):
            self.assertDenied(token)
        with self.subTest(mock_input_artifacts=True):
            self.assertAllowed(token, mock_input_artifacts=True)

    def test_public_with_user_token(self) -> None:
        """Try a user token on a public workspace."""
        token = self.playground.create_user_token()
        self.assertAllowed(token)

    def test_private_with_user_token(self) -> None:
        """Try a user token on a private workspace."""
        token = self.playground.create_user_token()
        self.make_private()
        self.assertDenied(token)

    def test_private_with_owner_token(self) -> None:
        """Try a owner user token on a private workspace."""
        token = self.playground.create_user_token()
        self.make_private()
        self.make_owner()
        self.assertAllowed(token)

    def test_public_with_token_disabled(self) -> None:
        """Try a disabled Token on a public workspace."""
        token = self.playground.create_user_token(enabled=False)
        self.assertAllowed(token)

    def test_private_with_token_disabled(self) -> None:
        """Try a disabled Token on a public workspace."""
        token = self.playground.create_user_token(enabled=False)
        self.make_private()
        self.assertDenied(token)

    def test_public_with_no_token(self) -> None:
        """Try without a token on a public workspace."""
        self.assertAllowed()

    def test_private_with_no_token(self) -> None:
        """Try without a token on a private workspace."""
        self.make_private()
        self.assertDenied()

    def test_public_with_user(self) -> None:
        """Try session user on a public workspace."""
        self.assertAllowed(user=self.user)

    def test_private_with_user(self) -> None:
        """Try session user on a private workspace."""
        self.make_private()
        self.assertDenied(user=self.user)

    def test_private_with_owner_user(self) -> None:
        """Try session user on an owned private workspace."""
        self.make_private()
        self.make_owner()
        self.assertAllowed(user=self.user)


del PermissionTests



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the WorkRequest UI helper."""

from django.test import RequestFactory

from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.ui.work_requests import WorkRequestUI


class WorkRequestUITests(TestCase):
    """Test the :py:class:`WorkRequestUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def ui(self, work_request: WorkRequest) -> WorkRequestUI:
        """Instantiate a UI helper for a workspace."""
        factory = RequestFactory()
        request = factory.get("/")
        return work_request.ui(request)

    def test_label_workflow_invalid_data(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        wr = self.playground.create_workflow(template, task_data={})
        assert wr.configured_task_data is not None
        wr.configured_task_data["invalid"] = True
        wr.save()
        self.assertEqual(self.ui(wr).label, "test")

    def test_label_for_synchronization_point(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        wr = WorkRequest.objects.create_synchronization_point(
            parent=workflow, step="test", display_name="Test"
        )
        self.assertEqual(self.ui(wr).label, "synchronization_point")

    def test_sidebar_work_request_status(self) -> None:
        wr = WorkRequest(status=WorkRequest.Statuses.PENDING)
        el = self.ui(wr).sidebar_work_request_status()
        assert el.value is not None
        self.assertHTMLEqual(
            el.value, '<span class="badge text-bg-secondary">Pending</span>'
        )

        wr = WorkRequest(status=WorkRequest.Statuses.BLOCKED)
        el = self.ui(wr).sidebar_work_request_status()
        assert el.value is not None
        self.assertHTMLEqual(
            el.value, '<span class="badge text-bg-dark">Blocked</span>'
        )

        wr.status, wr.result = (
            WorkRequest.Statuses.COMPLETED,
            WorkRequest.Results.FAILURE,
        )
        el = self.ui(wr).sidebar_work_request_status()
        assert el.value is not None
        self.assertHTMLEqual(
            el.value,
            '<span class="badge text-bg-primary">Completed</span>'
            ' <span class="badge text-bg-warning">Failure</span>',
        )

        wr.status, wr.result = (
            WorkRequest.Statuses.COMPLETED,
            WorkRequest.Results.SKIPPED,
        )
        el = self.ui(wr).sidebar_work_request_status()
        assert el.value is not None
        self.assertHTMLEqual(
            el.value,
            '<span class="badge text-bg-primary">Completed</span>'
            ' <span class="badge text-bg-secondary">Skipped</span>',
        )

        # This case does not make sense but it must be rendered somewhat
        wr.status, wr.result = (
            WorkRequest.Statuses.ABORTED,
            WorkRequest.Results.ERROR,
        )
        el = self.ui(wr).sidebar_work_request_status()
        assert el.value is not None
        self.assertHTMLEqual(
            el.value,
            '<span class="badge text-bg-dark">Aborted</span>'
            ' <span class="badge text-bg-danger">Error</span>',
        )

    def test_sidebar_supersede(self) -> None:
        wr_old = WorkRequest(pk=1, workspace=self.scenario.workspace)
        wr_new = WorkRequest(pk=2, workspace=self.scenario.workspace)
        wr_old.superseded = wr_new

        el = self.ui(wr_old).sidebar_work_request_superseded()
        self.assertEqual(el.value, "2")
        el = self.ui(wr_old).sidebar_work_request_supersedes()
        self.assertIsNone(el.value)
        el = self.ui(wr_new).sidebar_work_request_superseded()
        self.assertIsNone(el.value)
        el = self.ui(wr_new).sidebar_work_request_supersedes()
        self.assertEqual(el.value, "1")



# Generated by Django 4.2.21 on 2025-09-16 11:22

from django.db import migrations, models

import debusine.assets.models


class Migration(migrations.Migration):
    dependencies = [
        ('signing', '0003_register_pkcs11_static_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='auditlog',
            name='purpose',
            field=models.CharField(
                choices=[
                    (
                        debusine.assets.models.KeyPurpose['UEFI'],
                        debusine.assets.models.KeyPurpose['UEFI'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                    ),
                ],
                max_length=32,
            ),
        ),
        migrations.AlterField(
            model_name='key',
            name='purpose',
            field=models.CharField(
                choices=[
                    (
                        debusine.assets.models.KeyPurpose['UEFI'],
                        debusine.assets.models.KeyPurpose['UEFI'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                        debusine.assets.models.KeyPurpose['OPENPGP_UPLOAD'],
                    ),
                    (
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                        debusine.assets.models.KeyPurpose['OPENPGP_REPOSITORY'],
                    ),
                ],
                max_length=32,
            ),
        ),
    ]
